import numpy as np

# estimated from slide 34 of Karl's "Luminance 100 years" talk.
bright_JND = 1/1.65 # just take the inverse of the slope of the psychometric function
bright_JND = bright_JND/100 # since it was in units of 100% and we are typically working in [0,1] units
bright_JND

def maxRGB(rgb):
    """
    Returns the maximum value of the input RGB array.

    Parameters:
    rgb (numpy.ndarray): An array of RGB values.

    Returns:
    float: The maximum value of the input RGB array.
    """
    return np.max(rgb)


def maxWeightRGB(rgbs, wR, wG, wB):
    w_rgbs = np.array([[wR], [wG], [wB]]) * rgbs
    return np.array(np.max(w_rgbs, axis=0)).flatten()


def maxRGB_Shuchen(rgb):
    ws = np.array([0.40, 0.45, 0.15])
    return np.max(ws * rgb)


rand_weights = np.random.rand(1, 3)
def maxRGB_Random(rgb):
    return np.max(rand_weights * rgb)


def maxRGB_G0(rgb):
    r = rgb[0]
    g = rgb[1]
    b = rgb[2]

    return np.max([0.36593642*r,  0.42763098*g,  0.33796619*b])


import numpy as np

AliceWeight_mw = np.array([0.44, 0.46, 0.1])
AliceWeight_sw = np.array([0.41, 0.44, 0.15])

RGB_weights_aliceMonitor = np.array([43.8726816818746, 100.422336391086, 8.66744700006506])
RGB_weights_aliceMonitor = RGB_weights_aliceMonitor/RGB_weights_aliceMonitor.sum()

RGB_weights_radiance_aliceMonitor = np.array([0.334350089660000, 0.211504326128800, 0.143650257850000])
RGB_weights_radiance_aliceMonitor = RGB_weights_radiance_aliceMonitor/RGB_weights_radiance_aliceMonitor.sum()


import numpy as np

def sumRGB(rgb):
    """
    Sums the red, green, and blue values of an RGB color.

    Parameters:
    rgb (tuple): A tuple containing the red, green, and blue values of an RGB color.

    Returns:
    int: The sum of the red, green, and blue values.
    """
    return np.sum(rgb)


def sumWeightRGB(rgbs, wR, wG, wB):
    w_rgbs = np.array([[wR], [wG], [wB]]) * rgbs
    return np.array(np.sum(w_rgbs, axis=0)).flatten()


def sumRGB_Shuchen(rgb):
    ws = np.array([0.37, 0.45, 0.18])
    return np.sum(ws * rgb)


rand_weights = np.random.rand(1, 3)
def sumRGB_Random(rgb):
    return np.sum(rand_weights * rgb)


import numpy as np

newRW = 0.2667
newGW = 0.6267
newBW = 0.1067

def jingHSP(rgb, version='new'):
    r = rgb[0]
    g = rgb[1]
    b = rgb[2]

    N = 2.9

    t = 0
    if version == 'old':
        t = np.power(0.302*r, N) + np.power(0.550*g, N) + np.power(0.148*b, N)
    else:
        t = np.power(newRW*r, N) + np.power(newGW*g, N) + np.power(newBW*b, N)
    
    return np.power(t, 1/N)


import numpy as np

from .cie_monitor_helpers import rgb2xyz
from .monitor import Monitor

b = 1.15
g = 0.66
c1 = 3424/np.power(2, 12)
c2 = 2413/np.power(2, 7)
c3 = 2392/np.power(2, 7)
n = 2610/np.power(2, 14)
p = 1.7 * 2523/np.power(2, 5)
d = -0.56
d0 = 1.6295499532821566e-11

def xyz_corr(XYZ):
    X = XYZ[0]
    Y = XYZ[1]
    Z = XYZ[2]

    x_prime = b*X - (b-1)*Z
    y_prime = g*Y - (g-1)*X

    return np.array([x_prime, y_prime, Z])


adapt = np.array([[0.41478972, 0.579999, 0.014648],
                  [-0.20151,   1.120649, 0.0531008],
                  [-0.0166008, 0.2648,   0.6684799]])


def xyz_to_lms(xyz_corr):
    return adapt @ xyz_corr


def lms_prime(lms):
    a = c1 + c2*np.power(lms/10000, n)
    DD = 1 + c3*np.power(lms/10000, n)
    c = a/DD
    return np.power(c, p)


to_iaz = np.array([[0.5, 0.5, 0],
                   [3.524000, -4.066708, 0.542708],
                   [0.199076, 1.096799, -1.295875]])

def iab(lms_prime):
    return to_iaz @ lms_prime


def to_jaz(iab):
    I_z = iab[0]
    a_z = iab[1]
    b_z = iab[2]

    a = (1 + d)*I_z
    DD = 1 + d*I_z
    J_z = a/DD - d0

    return np.array([J_z, a_z, b_z])


def xyz_to_jaz(XYZ):
    """
    Convert a color from the XYZ color space to the Jazzy color space.

    Args:
        XYZ (tuple): A tuple containing the X, Y, and Z values of the color in the XYZ color space.

    Returns:
        tuple: A tuple containing the Jazzy color values.
    """
    return to_jaz(iab(lms_prime(xyz_to_lms(xyz_corr(XYZ)))))


def jazzy(mon: Monitor, rgb):
    """
    Converts an RGB color to Jazzy color space.

    Args:
        mon (Monitor): The monitor object.
        rgb (tuple): The RGB color to convert.

    Returns:
        tuple: The Jazzy color space representation of the RGB color.
    """
    xyz = rgb2xyz(mon, rgb)
    return xyz_to_jaz(xyz)


def jaz_chroma(jaz):
    """
    Calculates the chroma value of a given jaz array.

    Parameters:
    jaz (list): A list of length 3 containing the jaz values.

    Returns:
    float: The chroma value calculated as the square root of the sum of the squares of the second and third elements of the jaz array.
    """
    a = jaz[1]
    b = jaz[2]

    return np.sqrt(np.power(a, 2) + np.power(b, 2))


def jaz_hue(jaz):
    """
    Calculates the hue angle of a given color in the jazzy color space.

    Args:
        jaz (list): A list of three values representing the jazzy color space coordinates.

    Returns:
        float: The hue angle in radians.
    """
    a = jaz[1]
    b = jaz[2]

    theta = np.arctan(b/a)
    if theta < 0:
        theta += 2*np.pi

    return theta


def jaz_lightness(jaz):
    """
    Returns the lightness value of a given Jazzy color.

    Args:
        jaz (tuple): A tuple representing a Jazzy color in the format (lightness, hue, saturation).

    Returns:
        float: The lightness value of the Jazzy color.
    """
    return jaz[0]


def jaz_deltaE(jaz1, jaz2):
    """
    Calculates the color difference between two colors in the JAZ color space.

    Args:
        jaz1 (tuple): A tuple containing the L*, a*, and b* values of the first color.
        jaz2 (tuple): A tuple containing the L*, a*, and b* values of the second color.

    Returns:
        float: The color difference between the two colors.
    """
    J1 = jaz_lightness(jaz1)
    chroma1 = jaz_chroma(jaz1)
    hue1 = jaz_hue(jaz1)

    J2 = jaz_lightness(jaz2)
    chroma2 = jaz_chroma(jaz2)
    hue2 = jaz_hue(jaz2)

    deltaJ = J1 - J2
    deltaC = chroma1 - chroma2

    delta_little_h = hue1 - hue2
    deltaH = 2 * np.sqrt(chroma1*chroma2) * np.sin(delta_little_h/2)

    deltaE = np.sqrt(np.power(deltaJ, 2) + np.power(deltaC, 2) + np.power(deltaH, 2))

    return deltaE


import numpy as np
import colour

from .cie_standard import xyY2XYZ, XYZ2xy

to_lms = np.array([[0.38971, 0.68898, -0.07868],
                   [-0.22981, 1.18340, 0.04641],
                   [0.0, 0.0, 1.0]])

def xyz_to_cone(xyz):
    """
    Convert a set of XYZ coordinates to cone coordinates using the transformation matrix to_lms.

    Args:
        xyz (numpy.ndarray): A 3-element array representing the XYZ coordinates.

    Returns:
        numpy.ndarray: A 3-element array representing the cone coordinates.
    """
    return to_lms @ xyz


def f_n(I):
    a = np.power(I, 0.73)
    b = np.power(I, 0.73) + 2
    return 40*(a/b)


def adapt_cones(lms, lms_w, lms_p, La, Fl, Y_b, Y_white, discount, helson_judd):
    """
    Adapt cone responses to background illumination and discounting.

    Args:
    - lms (list): cone responses to the stimulus
    - lms_w (list): cone responses to the white point
    - lms_p (list): cone responses to the adapting stimulus
    - La (float): adapting luminance
    - Fl (float): luminance adaptation factor
    - Y_b (float): luminance of the background
    - Y_white (float): luminance of the white point
    - discount (bool): whether to discount the adapting stimulus
    - helson_judd (bool): whether to use the Helson-Judd effect

    Returns:
    - np.array: adapted cone responses
    """
    p = lms[0]
    g = lms[1]
    b = lms[2]

    p_w = lms_w[0]
    g_w = lms_w[1]
    b_w = lms_w[2]

    p_p = lms_p[0]
    g_p = lms_p[1]
    b_p = lms_p[2]

    t = p_w + g_w + b_w
    hp = 3 * p_w/t
    hg = 3 * g_w/t
    hb = 3 * b_w/t

    if not discount:
        Fp = (1 + np.power(La, 1/3) + hp)/(1 + np.power(La, 1/3) + 1/hp)
        Fg = (1 + np.power(La, 1/3) + hg)/(1 + np.power(La, 1/3) + 1/hg)
        Fb = (1 + np.power(La, 1/3) + hb)/(1 + np.power(La, 1/3) + 1/hb)
    else:
        Fp = 1.0
        Fg = 1.0
        Fb = 1.0

    fn_P = f_n(Fl * Fp * p/p_w)
    fn_G = f_n(Fl * Fg * g/g_w)
    fn_B = f_n(Fl * Fb * b/b_w)

    if helson_judd:
        r = Y_b/Y_white
        p_D = f_n(r * Fl * Fg) - f_n(r * Fl * Fp)
        g_D = 0.0
        b_D = f_n(r * Fl * Fg) - f_n(r * Fl * Fb)
    else:
        p_D = 0.0
        g_D = 0.0
        b_D = 0.0

    beta_p = np.power(10, 7)/(np.power(10, 7) + 5*La*(p_w/100))
    beta_g = np.power(10, 7)/(np.power(10, 7) + 5*La*(g_w/100))
    beta_b = np.power(10, 7)/(np.power(10, 7) + 5*La*(b_w/100))

    p_adapt = beta_p*(fn_P + p_D) + 1
    g_adapt = beta_g*(fn_G + g_D) + 1
    b_adapt = beta_b*(fn_B + b_D) + 1

    return np.array([p_adapt, g_adapt, b_adapt])


def achrom_cone(pgb):
    """
    Calculates the achromatic cone response for a given pixel.

    Args:
        pgb (tuple): A tuple containing the pixel values for the red, green, and blue channels.

    Returns:
        float: The achromatic cone response for the given pixel.
    """
    p = pgb[0]
    g = pgb[1]
    b = pgb[2]

    return 2*p + g + (1/20)*b - 3.05 + 1


def C(pgb):
    p = pgb[0]
    g = pgb[1]
    b = pgb[2]

    return np.array([
        p - g,
        g - b,
        b - p
    ])


def h_s(C):
    C1 = C[0]
    C2 = C[1]
    C3 = C[2]

    a = 0.5 * (C2 - C3)/4.5
    b = C1 - (C2/11)

    h = np.arctan2(a, b)
    if h < 0:
        h += 2*np.pi

    return np.rad2deg(h)


unique_red = {'hs': 20.14, 'es': 0.8}
unique_yellow = {'hs': 90.00, 'es': 0.7}
unique_green = {'hs': 164.25, 'es': 1.0}
unique_blue = {'hs': 237.53, 'es': 1.2}
def H(hs):
    # print("initial hs: " + str(hs))
    mhc = 0
    if hs > 0 and hs < unique_red['hs']:
        mhc = 3
        hs = hs + 360 # for proper es interpolation
    elif hs > unique_red['hs'] and hs < unique_yellow['hs']:
        mhc = 0
    elif hs > unique_yellow['hs'] and hs < unique_green['hs']:
        mhc = 1
    elif hs > unique_green['hs'] and hs < unique_blue['hs']:
        mhc = 2
    elif hs > unique_blue['hs'] and hs < unique_red['hs'] + 360:
        mhc = 3

    # rd = hs - unique_red['hs']
    # yd = hs - unique_yellow['hs']
    # gd = hs - unique_green['hs']
    # bd = hs - unique_blue['hs']

    # mhc = 0
    # hds = np.array([rd, yd, gd, bd])
    # print("hds: ", hds)
    # for hdc in range(len(hds)):
        # if hdc < 0:
            # mhc = hdc - 1
            # break
    
    # mhc = np.argmin(hds)
    # if hds[mhc] > 0:
        # mhc = mhc - 1
    
    # if mhc < 0:
        # mhc = 3

    if mhc == 0:
        H1 = 0

        h1 = unique_red['hs']
        e1 = unique_red['es']

        h2 = unique_yellow['hs']
        e2 = unique_yellow['es']
    elif mhc == 1:
        H1 = 100

        h1 = unique_yellow['hs']
        e1 = unique_yellow['es']

        h2 = unique_green['hs']
        e2 = unique_green['es']
    elif mhc == 2:
        H1 = 200

        h1 = unique_green['hs']
        e1 = unique_green['es']

        h2 = unique_blue['hs']
        e2 = unique_blue['es']
    elif mhc == 3:
        H1 = 300

        h1 = unique_blue['hs']
        e1 = unique_blue['es']

        h2 = unique_red['hs'] + 360 # for proper es interpolation
        e2 = unique_red['es']


    a = 100 * (hs - h1)/e1
    b = (hs - h1)/e1 + (h2 - hs)/e2

    hue = H1 + a/b

    a = (hs - h1)/(h2 - h1)
    es = (1 - a)*e1 + a*e2

    # uhues = np.array([
        # unique_red['hs'],
        # unique_yellow['hs'],
        # unique_green['hs'],
        # unique_blue['hs']
    # ])
    # ecens = np.array([
        # unique_red['es'],
        # unique_yellow['es'],
        # unique_green['es'],
        # unique_blue['es']
    # ])
    # es = np.interp(hs, uhues, ecens)

    # print("mhc: " + str(mhc))
    # print("hs: " + str(hs))
    # print("uhues: ", uhues)
    # print("ecens: ", ecens)

    # print("e1: " + str(e1))
    # print("e2: " + str(e2))
    # print("a: " + str(a))

    # print("es _ np: " + str(es))
    # print("es _ rob: " + str(es_rob))

    return hue, es


def M_YB(C, La, Nc, Ncb, es):
    C2 = C[1]
    C3 = C[2]

    Ft = La/(La + 0.1)

    a = 0.5 * (C2 - C3)/4.5
    b = es * (10/13) * Nc * Ncb * Ft
    
    return 100 * a * b


def M_RG(C, Nc, Ncb, es):
    C1 = C[0]
    C2 = C[1]

    a = C1 - (C2/11)
    b = es * (10/13) * Nc * Ncb

    return 100 * a * b


def M(M_YB, M_RG):
    return np.sqrt(np.power(M_YB, 2) + np.power(M_RG, 2))


def s(M, pgb):
    p = pgb[0]
    g = pgb[1]
    b = pgb[2]

    return 50 * M/(p + g + b)


def achrom_signal(Las, A, Nbb, S_S_w):
    j = 0.00001/(5*Las/2.26 + 0.00001)

    Fls = 3800 * np.power(j, 2) * (5*Las/2.26) + 0.2 * np.power(1 - np.power(j, 2), 4) * np.power(5*Las/2.26, 1/6)

    # should it be to power of 0.4? that is what is in the "colour" python package, but not the fairchild book
    # Fls = 3800 * np.power(j, 2) * (5*Las/2.26) + 0.2 * np.power(1 - np.power(j, 2), 0.4) * np.power(5*Las/2.26, 1/6)

    a = (5*Las/2.26)*S_S_w
    b = 1 + 5*(5*Las/2.26)
    c = 1 + 0.3*np.power(a, 0.3)
    Bs = 0.5/c + 0.5/b

    As = 3.05 * Bs * (f_n(Fls * S_S_w)) + 0.3

    return Nbb * (A - 1 + As - 0.3 + np.sqrt(1 + np.power(0.3, 2)))


def brightness(asig_a, asig_w, Nb, M):
    a = np.power(7*asig_w, 0.5)
    b = 5.33 * np.power(Nb, 0.13)
    N1 = a/b

    N2 = 7 * asig_w * np.power(Nb, 0.362)/200

    a = 7 * (asig_a + M/100)
    Q = np.power(a, 0.6) * N1 - N2
    
    return Q


def Q_wb(Q, Qb):
    return 20 * (np.power(Q, 0.7) - np.power(Qb, 0.7))


def lightness(Y_b, Y_w, Q, Q_w):
    Z = 1 + np.power(Y_b/Y_w, 0.5)
    return 100 * np.power(Q/Q_w, Z)
    

# values for Nc, Nb
# small area, uniform surround = 1.0, 300
# normal scenes = 1.0, 75
# television, dim surround = 1.0, 25
# transparencies, light box = 0.7, 25
# transparencies, dark surround = 0.7, 10
def hunt_brightness(XYZ_stim, XYZ_illum, La, XYZ_adapt, XYZ_bkgd, Nc, Nb, discount, helson_judd):
    """
    Calculates the brightness of a stimulus based on the Hunt model.

    Args:
    - XYZ_stim (tuple): tuple containing the tristimulus values of the stimulus in XYZ color space.
    - XYZ_illum (tuple): tuple containing the tristimulus values of the illuminant in XYZ color space.
    - La (float): adaptation luminance level in cd/m^2.
    - XYZ_adapt (tuple): tuple containing the tristimulus values of the adapting field in XYZ color space.
    - XYZ_bkgd (tuple): tuple containing the tristimulus values of the background in XYZ color space.
    - Nc (float): surround luminance level in cd/m^2.
    - Nb (float): background luminance level in cd/m^2.
    - discount (float): discounting factor.
    - helson_judd (bool): whether to use Helson-Judd effect or not.

    Returns:
    - dict: a dictionary containing the following keys:
        - 'Q': the brightness of the stimulus.
        - 'WB': the brightness of the stimulus with respect to the background.
        - 'J': the lightness of the stimulus.
    """
    xy_illum = XYZ2xy(XYZ_illum)

    xyz = XYZ_stim
    Y_stim = XYZ_stim[1]

    xyz_w = XYZ_illum
    Y_white = XYZ_illum[1]

    xyz_bkg = XYZ_bkgd
    Y_b = XYZ_bkgd[1]

    xyz_prox = XYZ_bkgd
    Y_p = xyz_prox[1]

    T = colour.temperature.xy_to_CCT(xy_illum, "Hernandez 1999")

    Las = 2.26*La*np.power(T/4000 - 0.4, 1/3)
    k = 1/(5*La + 1)
    Fl = 0.2 * np.power(k, 4) * (5*La) + 0.1*np.power(1 - np.power(k, 4), 2) * np.power(5*La, 1/3)

    Ncb = 0.725*np.power(Y_white/Y_b, 0.2)
    Nbb = 0.725*np.power(Y_white/Y_b, 0.2)

    # white (cones)
    pgb_w = xyz_to_cone(xyz_w)

    # proximal
    pgb_p = xyz_to_cone(xyz_prox)

    # stimulus
    pgb = xyz_to_cone(xyz)
    pgb_a = adapt_cones(pgb, pgb_w, pgb_p, La, Fl, Y_b, Y_white, discount, helson_judd)
    Aa = achrom_cone(pgb_a)
    Ca = C(pgb_a)
    hue, es = H(h_s(Ca))
    M_YBa = M_YB(Ca, La, Nc, Ncb, es)
    M_RGa = M_RG(Ca, Nc, Ncb, es)
    Ma = M(M_YBa, M_RGa)
    sa = s(Ma, pgb_a)
    S_S_w = Y_stim/Y_white

    asig_a = achrom_signal(Las, Aa, Nbb, S_S_w)

    # white
    pgb_aw = adapt_cones(pgb_w, pgb_w, pgb_p, La, Fl, Y_b, Y_white, discount, helson_judd)
    Aw = achrom_cone(pgb_aw)
    Caw = C(pgb_aw)
    # hue, es = H(h_s(Caw))
    M_YBaw = M_YB(Caw, La, Nc, Ncb, es)
    M_RGaw = M_RG(Caw, Nc, Ncb, es)
    Maw = M(M_YBaw, M_RGaw)
    S_w_S_w = Y_white/Y_white

    asig_w = achrom_signal(Las, Aw, Nbb, S_w_S_w)

    Qa = brightness(asig_a, asig_w, Nb, Ma)
    Qw = brightness(asig_w, asig_w, Nb, Maw)

    # background
    pgb_bkg = xyz_to_cone(xyz_bkg)
    pgb_bkg_a = adapt_cones(pgb_bkg, pgb_w, pgb_p, La, Fl, Y_b, Y_white, discount, helson_judd)
    Aa_bkg = achrom_cone(pgb_bkg_a)
    Ca_bkg = C(pgb_bkg_a)
    hue_bkg, es_bkg = H(h_s(Ca_bkg))
    M_YBa_bkg = M_YB(Ca_bkg, La, Nc, Ncb, es_bkg)
    M_RGa_bkg = M_RG(Ca_bkg, Nc, Ncb, es_bkg)
    Ma_bkg = M(M_YBa_bkg, M_RGa_bkg)
    sa_bkg = s(Ma_bkg, pgb_bkg_a)

    S_bkg_S_w = Y_b/Y_white
    
    asig_bkg = achrom_signal(Las, Aa_bkg, Nbb, S_bkg_S_w)
    Qa_bkg = brightness(asig_bkg, asig_w, Nb, Ma_bkg)

    Q_wba = Q_wb(Qa, Qa_bkg)

    J = lightness(Y_b, Y_white, Qa, Qw)

    return {
        'Q': Qa,
        'WB': Q_wba,
        'J': J
    }


import numpy as np
from scipy.interpolate import CubicSpline

from .monitor import Monitor

konica_wlns = np.arange(380, 781)

wlns = np.array([420,
                 430,
                 440,
                 450,
                 459,
                 460,
                 470,
                 480,
                 490,
                 500,
                 510,
                 520,
                 526,
                 530,
                 540,
                 550,
                 560,
                 570,
                 580,
                 590,
                 600,
                 610,
                 620,
                 630,
                 640,
                 645,
                 650,
                 660,
                 670,
                 680])

A = np.array([-0.140,
              -0.240,
              -0.360,
              -0.470,
              -0.520,
              -0.520,
              -0.500,
              -0.460,
              -0.400,
              -0.360,
              -0.310,
              -0.260,
              -0.230,
              -0.200,
              -0.140,
              -0.080,
               0.020,
               0.200,
               0.460,
               0.700,
               0.830,
               0.890,
               0.920,
               0.940,
               0.955,
               0.960,
               0.965,
               0.970,
               0.975,
               0.980])
A_interp = CubicSpline(wlns, A)
A_adj = A_interp(konica_wlns)

B = np.array([0.640,
              0.680,
              0.800,
              0.850,
              0.880,
              0.880,
              0.890,
              0.910,
              0.920,
              0.930,
              0.950,
              0.955,
              0.960,
              0.965,
              0.970,
              0.970,
              0.965,
              0.950,
              0.860,
              0.710,
              0.550,
              0.460,
              0.400,
              0.340,
              0.310,
              0.290,
              0.280,
              0.245,
              0.215,
              0.190])
B_interp = CubicSpline(wlns, B)
B_adj = B_interp(konica_wlns)

C = np.array([0.710,
              0.650,
              0.460,
              0.250,
              0.070,
              0.070,
             -0.080,
             -0.160,
             -0.220,
             -0.240,
             -0.250,
             -0.260,
             -0.260,
             -0.260,
             -0.260,
             -0.250,
             -0.235,
             -0.200,
             -0.150,
             -0.070,
             -0.010,
              0.030,
              0.060,
              0.075,
              0.080,
              0.085,
              0.090,
              0.095,
              0.100,
              0.110])
C_interp = CubicSpline(wlns, C)
C_adj = C_interp(konica_wlns)

from .cmfs import V_lambda

# assume for now that all spectra were recorded with the konica cs2000-a
def guth_lum_spect(spect):
    """
    Calculates the Guth luminosity of a given spectrum.

    Parameters:
    spect (numpy.ndarray): The input spectrum.

    Returns:
    float: The Guth luminosity of the input spectrum.
    """
    spect_lum = V_lambda * spect

    As = np.power(np.sum(spect * A_adj), 2.0)
    Bs = np.power(np.sum(spect * B_adj), 2.0)
    Cs = np.power(np.sum(spect * C_adj), 2.0)

    return np.sqrt(As + Bs + Cs)


def guth_lum_rgb(mon: Monitor, rgb):
    """
    Calculates the Guth luminance for a given RGB color value.

    Args:
        mon (Monitor): The monitor object containing the maximum spectral values for each color channel.
        rgb (tuple): A tuple containing the red, green, and blue values for the color.

    Returns:
        float: The Guth luminance value for the given RGB color.
    """
    r = rgb[0]
    g = rgb[1]
    b = rgb[2]

    R_spect = r * mon.R_max_spectrum
    G_spect = g * mon.G_max_spectrum
    B_spect = b * mon.B_max_spectrum

    spect = R_spect + G_spect + B_spect

    return guth_lum_spect(spect)


import numpy as np

from .monitor import Monitor
from .infamous_lab import rgb2lab

def lab2lch(lab):
    """
    Convert a color from CIELAB to CIELCH color space.

    Parameters:
    lab (numpy.ndarray): An array of shape (3,) representing the CIELAB color.

    Returns:
    numpy.ndarray: An array of shape (3,) representing the CIELCH color.
    """
    L = lab[0]
    a = lab[1]
    b = lab[2]

    C = np.sqrt(np.power(a, 2) + np.power(b, 2))

    H = np.arctan2(b, a)
    if H < 0:
        H = H + 2*np.pi

    return np.array([L, C, H])


def rgb2lch(mon: Monitor, rgb):
    """
    Convert RGB color space to LCH color space.

    Args:
        mon (Monitor): The monitor object.
        rgb (tuple): The RGB color values.

    Returns:
        tuple: The LCH color values.
    """
    lab = rgb2lab(mon, rgb)
    return lab2lch(lab)


import numpy as np
import scipy.io as sio
from scipy.interpolate import CubicSpline

from .cie_standard import xyY2XYZ
from .cmfs import lms_wlns, l_absorp, m_absorp, s_absorp
from .dkl_setup import cie2lms, lumchrm, solvex, solvey
# from .uv_space import xyY2uvY

class Monitor:
    def __init__(self, name):
        self.name = name

        self.monxyY_file = ""
        self.spectralData_file = ""

        self.monxyY = np.array([])
        self.monXYZ = np.array([])
        # self.monuvY = np.array([])
        self.monWP = np.array([])
        self.maxLuminance = np.array([])

        # default assumption is 2.2 gamma for all primaries
        self.monGamma = np.array([2.2, 2.2, 2.2])

        # our Konica CS2000-a measures in range of 380 to 780nm
        # in steps of 1nm
        self.wlns = np.arange(380, 781)

        self.spectralData = {}

        self.R_max_spectrum = np.array([])
        self.G_max_spectrum = np.array([])
        self.B_max_spectrum = np.array([])
        self.W_max_spectrum = np.array([])

        self.R_max_luminance = -1
        self.G_max_luminance = -1
        self.B_max_luminance = -1
        self.W_max_luminance = -1

        self.RGB2LMS = np.array([])
        self.LMS2RGB = np.array([])
        self.DKL2RGB = np.array([])
        self.RGB2DKL = np.array([])

        self.XYZ2RGB = np.array([])
        self.RGB2XYZ = np.array([])

        self.visibleRadianceWindow = np.zeros((len(self.wlns)))

    def load_monxyY(self, fname):
        self.monxyY_file = fname
        self.monxyY = np.genfromtxt(fname, delimiter=',')
        self.calc_monXYZ()

    def set_monxyY(self, monxyY):
        self.monxyY = monxyY
        self.calc_monXYZ()

    def calc_monXYZ(self):
        self.monXYZ = np.zeros((3, 3))
        for c in range(3):
            self.monXYZ[:, c] = xyY2XYZ(self.monxyY[c, :])

        self.monWP = self.monXYZ.sum(axis = 1)
        self.maxLuminance = self.monWP[1]

        # self.monuvY = np.zeros(self.monxyY.shape)
        # for c in range(3):
            # self.monuvY[c, :] = xyY2uvY(self.monxyY[c, :])

        self.RGB2XYZ = self.monXYZ
        self.XYZ2RGB = np.linalg.inv(self.RGB2XYZ)

        self.dkl2rgbFromCalib()

    def set_monGamma(self, monGamma):
        self.monGamma = monGamma

    def load_spectralData(self, fname):
        self.spectralData_file = fname

        self.spectralData = sio.loadmat(fname)

        self.R_max_spectrum = (self.spectralData['SPECTRA'][0][0]['spectralData'][0][0]).flatten()
        self.R_max_luminance = self.spectralData['SPECTRA'][0][0]['Y'][0][0].flatten()[0]

        self.G_max_spectrum = (self.spectralData['SPECTRA'][0][1]['spectralData'][0][0]).flatten()
        self.G_max_luminance = self.spectralData['SPECTRA'][0][1]['Y'][0][0].flatten()[0]

        self.B_max_spectrum = (self.spectralData['SPECTRA'][0][2]['spectralData'][0][0]).flatten()
        self.B_max_luminance = self.spectralData['SPECTRA'][0][2]['Y'][0][0].flatten()[0]

        self.W_max_spectrum = (self.spectralData['SPECTRA'][0][3]['spectralData'][0][0]).flatten()
        self.W_max_luminance = self.spectralData['SPECTRA'][0][3]['Y'][0][0].flatten()[0]

        self.calc_spectralQuants()

    def set_R_spectrum(self, wlns, R_max_spectrum, R_max_luminance):
        r_interp = CubicSpline(wlns, R_max_spectrum)
        self.R_max_spectrum = r_interp(self.wlns)
        self.R_max_spectrum[self.R_max_spectrum < 0] = 0

        self.R_max_luminance = R_max_luminance

    def set_G_spectrum(self, wlns, G_max_spectrum, G_max_luminance):
        g_interp = CubicSpline(wlns, G_max_spectrum)
        self.G_max_spectrum = g_interp(self.wlns)
        self.G_max_spectrum[self.G_max_spectrum < 0] = 0

        self.G_max_luminance = G_max_luminance

    def set_B_spectrum(self, wlns, B_max_spectrum, B_max_luminance):
        b_interp = CubicSpline(wlns, B_max_spectrum)
        self.B_max_spectrum = b_interp(self.wlns)
        self.B_max_spectrum[self.B_max_spectrum < 0] = 0

        self.B_max_luminance = B_max_luminance

    def set_W_spectrum(self, wlns, W_max_spectrum, W_max_luminance):
        w_interp = CubicSpline(wlns, W_max_spectrum)
        self.W_max_spectrum = w_interp(self.wlns)
        self.W_max_spectrum[self.W_max_spectrum < 0] = 0

        self.W_max_luminance = W_max_luminance

    def calc_spectralQuants(self):
        self.maxRadiance = self.W_max_spectrum.sum()

        self.rgb2lmsTransform()
        self.computeVisibleRadianceWindow()

        self.visible_R_max_spectrum = self.R_max_spectrum * self.visibleRadianceWindow
        self.visible_G_max_spectrum = self.G_max_spectrum * self.visibleRadianceWindow
        self.visible_B_max_spectrum = self.B_max_spectrum * self.visibleRadianceWindow

        self.visible_W_max_spectrum = self.W_max_spectrum * self.visibleRadianceWindow

        self.maxVisibleRadiance = self.visible_W_max_spectrum.sum()

    def rgb2lmsTransform(self):
        l_interp = CubicSpline(lms_wlns, l_absorp)
        self.l_adj = l_interp(self.wlns)
        self.l_adj[self.l_adj < 0] = 0

        m_interp = CubicSpline(lms_wlns, m_absorp)
        self.m_adj = m_interp(self.wlns)
        self.m_adj[self.m_adj < 0] = 0

        # s curve stops early, so correct all the NaNs to 0, as they should be
        s_absorp[np.isnan(s_absorp)] = 0

        s_interp = CubicSpline(lms_wlns, s_absorp)
        self.s_adj = s_interp(self.wlns)
        self.s_adj[self.s_adj < 0] = 0

        # offs1 = np.arange(11, len(rs))
        # offs2 = np.arange(1, len(rs) - 10)

        rs = self.R_max_spectrum
        gs = self.G_max_spectrum
        bs = self.B_max_spectrum

        rL = rs @ self.l_adj
        rM = rs @ self.m_adj
        rS = rs @ self.s_adj

        gL = gs @ self.l_adj
        gM = gs @ self.m_adj
        gS = gs @ self.s_adj

        bL = bs @ self.l_adj
        bM = bs @ self.m_adj
        bS = bs @ self.s_adj

        self.RGB2LMS = np.array([
            [rL, gL, bL],
            [rM, gM, bM],
            [rS, gS, bS]])
        self.LMS2RGB = np.linalg.inv(self.RGB2LMS)

    def computeVisibleRadianceWindow(self):
        l_max_idx = np.argmax(self.l_adj)
        s_max_idx = np.argmax(self.s_adj)

        for c in range(s_max_idx+1):
            self.visibleRadianceWindow[c] = self.s_adj[c]

        for c in range(s_max_idx+1, l_max_idx+1):
            self.visibleRadianceWindow[c] = 1.0

        for c in range(l_max_idx+1, len(self.visibleRadianceWindow)):
            self.visibleRadianceWindow[c] = self.l_adj[c]

    # from Qasim Zaidi
    def dkl2rgbFromCalib(self):
        white = np.zeros((3,))

        rx = self.monxyY[0, 0]
        ry = self.monxyY[0, 1]
        white[0] = self.monxyY[0, 2]/2

        gx = self.monxyY[1, 0]
        gy = self.monxyY[1, 1]
        white[1] = self.monxyY[1, 2]/2
        
        bx = self.monxyY[2, 0]
        by = self.monxyY[2, 1]
        white[2] = self.monxyY[2, 2]/2

        R = cie2lms(rx, ry)
        G = cie2lms(gx, gy)
        B = cie2lms(bx, by)

        L = np.array([R[0], G[0], B[0]])
        M = np.array([R[1], G[1], B[1]])
        S = np.array([R[2], G[2], B[2]])

        lumchrm(white[0], white[1], white[2], L, M, S)

        print('Red Green Axis')
        deltaGrg = solvex(white[0]*S[0], white[0]*(L[0]+M[0]), S[1], S[2], L[1]+M[1], L[2]+M[2])
        deltaBrg = solvey(white[0]*S[0], white[0]*(L[0]+M[0]), S[1], S[2], L[1]+M[1], L[2]+M[2])
        dGrg = -1*deltaGrg/white[1]
        dBrg = -1*deltaBrg/white[2]
        lumchrm(0.0, white[1]+deltaGrg, white[2]+deltaBrg, L, M, S)
        lumchrm(white[0]*2.0, white[1]-deltaGrg, white[2]-deltaBrg, L, M, S)

        print('Blue Yellow Axis')
        deltaRyv = solvex(white[2]*L[2], white[2]*M[2], L[0], L[1], M[0], M[1])
        deltaGyv = solvey(white[2]*L[2], white[2]*M[2], L[0], L[1], M[0], M[1])
        dRyv = -1*deltaRyv/white[0]
        dGyv = -1*deltaGyv/white[1]
        lumchrm(white[0]+deltaRyv, white[1]+deltaGyv, 0.0, L, M, S)
        lumchrm(white[0]-deltaRyv, white[1]-deltaGyv, white[2]*2.0, L, M, S)

        self.DKL2RGB = np.array([
                        [1, 1,    dRyv],
                        [1, dGrg, dGyv],
                        [1, dBrg, 1]])
        self.RGB2DKL = np.linalg.inv(self.DKL2RGB)


import numpy as np
import scipy.io as sio
from scipy.optimize import curve_fit

inds = [0, 32, 64, 83, 128, 160, 192, 220, 255]

def gamma(x, g, a):
    """
    Calculate the gamma function for a given input x, power g, and coefficient a.

    Args:
        x (float): Input value.
        g (float): Power value.
        a (float): Coefficient value.

    Returns:
        float: Result of the gamma function.
    """
    return a*np.power(x, g)


def fit_gamma(fn):
    """
    Fits gamma curves to the red, green, and blue channels of an image.

    Args:
        fn (str): The filename of the image to fit gamma curves to.

    Returns:
        None
    """
    lum_mat = sio.loadmat(fn)
    
    rys = np.array(lum_mat["rys"])
    gys = np.array(lum_mat["gys"])
    bys = np.array(lum_mat["bys"])

    gams = []
    gains = []
    for fc in range(rys.shape[0]):
        params, pcov = curve_fit(gamma, np.array(inds)/255, rys[fc]/np.max(rys[fc]))
        rg = params[0]
        rga = params[1]
        params, pcov = curve_fit(gamma, np.array(inds)/255, gys[fc]/np.max(gys[fc]))
        gg = params[0]
        gga = params[1]
        params, pcov = curve_fit(gamma, np.array(inds)/255, bys[fc]/np.max(bys[fc]))
        bg = params[0]
        bga = params[1]

        gams.append([rg, gg, bg])
        gains.append([rga, gga, bga])

    sio.savemat(fn, {"gammas": np.array(gams), "gains": np.array(gains)})


# def konica_spect_to_xyz(spect):
#     wlns = np.arange(380, 780)
#     cmfData = csvread('ciexyz31_1.csv')

#     wavelength_cmf = cmfData[:, 0]
#     x_bar = cmfData[:, 1]
#     y_bar = cmfData[:, 2]
#     z_bar = cmfData[:, 3]

#     cmf = np.interp(wavelength_cmf, [x_bar, y_bar, z_bar], wlns, 'spline')
#     cmf[cmf < 0] = 0

#     XYZ = np.zeros((3, 1))

#     stepSize = [diff(wlns), 0]

#     corrCMF[:, 0] = 683 * cmf[:, 0] * stepSize
#     corrCMF[:, 1] = 683 * cmf[:, 1] * stepSize
#     corrCMF[:, 2] = 683 * cmf[:, 2] * stepSize

#     return spect.T @ corrCMF


import numpy as np

atd_jnd_macadam = 0.002
atd_jnd = 0.005 # page 243 of fairchild book, guth - "wide variety of color discrim phenomena"

# small diffs are quadrature sum of del_A1, del_T1, and del_D1
# large diffs are quadrature sum of del_A2, del_T2, and del_D2

def vos_corr_xyz(x, y, z):
    a = 1.027*x - 0.00008*y - 0.0009
    b = 0.03845*x + 0.01496*y + 1
    xp = a/b

    c = 0.00376*x + 1.0072*y + 0.00764
    d = 0.03845*x + 0.0149*y + 1
    yp = c/d

    zp = 1 - xp - yp

    return np.array([xp, yp, zp])


def xyz_to_lms(Xp, Yp, Zp):
    """
    Convert from XYZ color space to LMS color space using the ATD95 transformation.

    Args:
        Xp (float): X component of the color in XYZ color space.
        Yp (float): Y component of the color in XYZ color space.
        Zp (float): Z component of the color in XYZ color space.

    Returns:
        numpy.ndarray: Array containing the L, M, and S components of the color in LMS color space.
    """
    l = np.power(0.66 * (0.2435 * Xp + 0.8524 * Yp - 0.0516 * Zp), 0.7) + 0.024
    m = np.power(1.0 * (-0.3954 * Xp + 1.1642 * Yp + 0.0837 * Zp), 0.7) + 0.036
    s = np.power(0.43 * (0.04 * Yp + 0.6225 * Zp), 0.7) + 0.31

    return np.array([l, m, s])


# XYZ values should be Judd modified values
# XYZ adapt is normally assumed to be perfect white under the illumination
# for unrelated colors, k1 = 1.0 and k2 = 0.0
# for related colors, k1 = 0.0 and k2 is btween 15 and 50
# if adapted to stimulus and white point, then k1 = 1.0 and k2 = 5.0
# if we want a "real" von Kries transform, then  k1 = 0.0 and k2 = 50
def guth_atd(xyz_stim, xyz_adapt, Y_illum, k1, k2):
    """
    Computes the ATD (Adaptation-Test-Distance) color space values for a given stimulus and adaptation condition.

    Args:
    - xyz_stim (list): A list of 3 floats representing the tristimulus values of the stimulus.
    - xyz_adapt (list): A list of 3 floats representing the tristimulus values of the adaptation condition.
    - Y_illum (float): The luminance of the illuminant.
    - k1 (float): A scaling factor for the stimulus tristimulus values.
    - k2 (float): A scaling factor for the adaptation tristimulus values.

    Returns:
    - A dictionary containing the following ATD color space values:
        - A1 (float): First stage A response.
        - T1 (float): First stage T response.
        - D1 (float): First stage D response.
        - A2 (float): Second stage A response.
        - T2 (float): Second stage T response.
        - D2 (float): Second stage D response.
        - Br (float): Brightness.
        - H (float): Hue.
        - C (float): Saturation.
    """
    X = xyz_stim[0]
    Y = xyz_stim[1]
    Z = xyz_stim[2]

    s = X + Y + Z
    x = X/s
    y = Y/s
    z = Z/s
    vos_xyz = vos_corr_xyz(x, y, z)

    r = Y/vos_xyz[1]
    Xp = 18 * np.power(vos_xyz[0]*r, 0.8)
    Yp = 18 * np.power(vos_xyz[1]*r, 0.8)
    Zp = 18 * np.power(vos_xyz[2]*r, 0.8)

    lms = xyz_to_lms(Xp, Yp, Zp)
    l = lms[0]
    m = lms[1]
    s = lms[2]

    Xo = xyz_adapt[0]
    Yo = xyz_adapt[1]
    Zo = xyz_adapt[2]

    so = Xo + Yo + Zo
    xo = Xo/so
    yo = Yo/so
    zo = Zo/so
    vos_xoyozo = vos_corr_xyz(xo, yo, zo)

    r = Yo/vos_xoyozo[1]
    Xop = 18 * np.power(vos_xoyozo[0]*r, 0.8)
    Yop = 18 * np.power(vos_xoyozo[1]*r, 0.8)
    Zop = 18 * np.power(vos_xoyozo[2]*r, 0.8)

    Xa = k1 * Xp + k2 * Xop
    Ya = k1 * Yp + k2 * Yop
    Za = k1 * Zp + k2 * Zop

    lms_a = xyz_to_lms(Xa, Ya, Za)
    la = lms_a[0]
    ma = lms_a[1]
    sa = lms_a[2]

    # sig is normally 300, but can be varied to fit data better
    sig = 300

    lg = l * (sig/(sig + la))
    mg = m * (sig/(sig + ma))
    sg = s * (sig/(sig + sa))

    # first stage initial signals
    A1i = 3.57 * lg + 2.64 * mg
    T1i = 7.18 * lg - 6.21 * mg
    D1i = -0.70 * lg + 0.085 * mg + 1 * sg

    # second stage initial signals
    A2i = 0.09 * A1i
    T2i = 0.43 * T1i + 0.76 * D1i
    D2i = D1i

    # final atd responses
    A1 = A1i/(200 + np.abs(A1i))
    T1 = T1i/(200 + np.abs(T1i))
    D1 = D1i/(200 + np.abs(D1i))

    A2 = A2i/(200 + np.abs(A2i))
    T2 = T2i/(200 + np.abs(T2i))
    D2 = D2i/(200 + np.abs(D2i))

    brightness = np.sqrt(np.power(A1, 2) + np.power(T1, 2) + np.power(D1, 2))
    saturation = np.sqrt(np.power(T2, 2) + np.power(D2, 2))/A2

    # in guth's original formulation hue = T2/D2, but this does not work in practice
    hue = np.arctan2(T2, D2)
    if hue < 0:
        hue += 2*np.pi
    
    hue = np.rad2deg(hue)
    # hue = T2/D2

    return {
        "A1": A1,
        "T1": T1,
        "D1": D1,
        "A2": A2,
        "T2": T2,
        "D2": D2,
        "Br": brightness,
        "H": hue,
        "C": saturation
    }


import numpy as np
import colour

from .nayatani_insanity import robNT
from .cie_standard import XYZ2xy

# Table 11.1 Example Nayatani et al. color appearance model calculations
# Quantity Case 1 Case 2 Case 3 Case 4
Xs = [19.01, 57.06, 3.53, 19.01]
Ys = [20.00, 43.06, 6.56, 20.00]
Zs = [21.78, 31.96, 2.14, 21.78]
Xns = [95.05, 95.05, 109.85, 109.85]
Yns = [100.00, 100.00, 100.00, 100.00]
Zns = [108.88, 108.88, 35.58, 35.58]
Eos = [5000, 500, 5000, 500]
Eors = [1000, 1000, 1000, 1000]
Brs = [62.6, 67.3, 37.5, 44.2]
Lps = [50.0, 73.0, 24.5, 49.4]
Lns = [50.0, 75.9, 29.7, 49.4]
thetas = [257.5, 21.6, 190.6, 236.3]
Hs = [317.8, 2.1, 239.4, 303.6]
Hcs = ["82B 18R", "98R 2Y", "61G 39B", "96B 4R"]
Ss = [0.0, 37.1, 81.3, 40.2]
Cs = [0.0, 48.3, 49.3, 39.9]
Ms = [0.0, 42.9, 62.1, 35.8]

for tc in range(4):
    X = Xs[tc]
    Y = Ys[tc]
    Z = Zs[tc]

    Xn = Xns[tc]
    Yn = Yns[tc]
    Zn = Zns[tc]

    Eo = Eos[tc]
    Eor = Eors[tc]

    xy_stim = XYZ2xy(np.array([X, Y, Z]))
    Y_stim = Y/Eo
    # Y_stim = Y/Yn
    # Yo_bg = Yn/Eo
    Yo_bg = 0.18
    xy_illum = XYZ2xy(np.array([Xn, Yn, Zn]))

    res = robNT(xy_stim, Y_stim, Yo_bg, xy_illum, Eo, Eor)
    res_Br = res['Brightness']
    res_Lp = res['Lightness']
    res_Ln = res['Normalized Lightness']
    res_theta = res['Hue Angle']
    res_S = res['Saturation']
    res_C = res['Chroma']
    res_M = res['Colorfulness']

    XYZ = np.array([X, Y, Z])
    XYZ_n = np.array([Xn, Yn, Zn])
    Y_o = 0.18

    res_col = colour.appearance.nayatani95.XYZ_to_Nayatani95(XYZ, XYZ_n, Y_o, Eo, Eor)
    # res_col_Br = res_col.B_r
    res_col_Q = res_col.Q
    res_col_Lp = res_col.L_star_P
    res_col_Ln = res_col.L_star_N
    res_col_theta = res_col.h
    res_col_S = res_col.s
    res_col_C = res_col.C
    res_col_M = res_col.M

    if Brs[tc] != res_Br:
        print(tc, "Brs: ", Brs[tc], ", Rob: ", res_Br, "Colour: ", res_col_Q)

    if Lps[tc] != res_Lp:
        print(tc, "Lps: ", Lps[tc], ", Rob: ", res_Lp, "Colour: ", res_col_Lp)

    if Lns[tc] != res_Ln:
        print(tc, "Lns: ", Lns[tc], ", Rob: ", res_Ln, "Colour: ", res_col_Ln)

    if thetas[tc] != res_theta:
        print(tc, "thetas: ", thetas[tc], ", Rob: ", res_theta, "Colour: ", res_col_theta)

    if Ss[tc] != res_S:
        print(tc, "Ss: ", Ss[tc],  ", Rob: ",res_S, "Colour: ", res_col_S)

    if Cs[tc] != res_C:
        print(tc, "Cs: ", Cs[tc],  ", Rob: ",res_C, "Colour: ", res_col_C)

    if Ms[tc] != res_M:
        print(tc, "Ms: ", Ms[tc],  ", Rob: ",res_M, "Colour: ", res_col_M)


import numpy as np

from .cie_standard import plank_lasso, XYZ2xyY
from .monitor import Monitor

def to_u(XYZ):
    """
    Converts a 3D point in XYZ space to its corresponding U coordinate in UV space.

    Args:
        XYZ (tuple): A tuple containing the X, Y, and Z coordinates of the point in XYZ space.

    Returns:
        float: The U coordinate of the point in UV space.
    """
    X = XYZ[0]
    Y = XYZ[1]
    Z = XYZ[2]

    return (4*X) / (X + 15*Y + 3*Z)


def to_v(XYZ):
    """
    Converts a 3D point in XYZ space to its corresponding V coordinate in UV space.

    Args:
        XYZ (tuple): A tuple containing the X, Y, and Z coordinates of the point in XYZ space.

    Returns:
        float: The V coordinate of the point in UV space.
    """
    X = XYZ[0]
    Y = XYZ[1]
    Z = XYZ[2]

    return (9*Y) / (X + 15*Y + 3*Z)


def xyY2uvY(xyY):
    """
    Convert xyY color space to uvY color space.

    Parameters:
    xyY (numpy.ndarray): Array of xyY values.

    Returns:
    numpy.ndarray: Array of uvY values.
    """
    x = xyY[0]
    y = xyY[1]
    Y = xyY[2]

    u = (4*x) / (-2*x + 12*y + 3)
    v = (9*y) / (-2*x + 12*y + 3)

    return np.array([u, v, Y])


def uvY2xyY(uvY):
    """
    Convert uvY color space to xyY color space.

    Args:
        uvY (np.array): Array of length 3 containing u, v, and Y values.

    Returns:
        np.array: Array of length 3 containing x, y, and Y values.
    """
    u = uvY[0]
    v = uvY[1]
    Y = uvY[2]

    x = (9*u) / (6*u - 16*v + 12)
    y = (4*v) / (6*u - 16*v + 12)
    
    return np.array([x, y, Y])


def XYZ2LUV(mon: Monitor, XYZ):
    """
    Convert XYZ color space to LUV color space.

    Args:
        mon (Monitor): The monitor object.
        XYZ (numpy.ndarray): The input color in XYZ color space.

    Returns:
        numpy.ndarray: The color in LUV color space.
    """
    Y = XYZ[1]
    Yn = mon.monWP[1]

    delta = 6/29
    L = 0
    if Y/Yn <= np.power(delta, 3):
        L = np.power(29/3, 3) * Y/Yn
    else:
        L = 116 * np.power(Y/Yn, 1/3) - 16

    un = to_u(mon.monWP)
    vn = to_v(mon.monWP)

    u = to_u(XYZ)
    v = to_v(XYZ)

    u_star = 13*L*(u - un)
    v_star = 13*L*(v - vn)

    return np.array([L, u_star, v_star])


def XYZ2LUV_wo_mon(XYZ, WP):
    Y = XYZ[1]
    Yn = WP[1]

    delta = 6/29
    L = 0
    if Y/Yn <= np.power(delta, 3):
        L = np.power(29/3, 3) * Y/Yn
    else:
        L = 116 * np.power(Y/Yn, 1/3) - 16

    un = to_u(WP)
    vn = to_v(WP)

    u = to_u(XYZ)
    v = to_v(XYZ)

    u_star = 13*L*(u - un)
    v_star = 13*L*(v - vn)

    return np.array([L, u_star, v_star])


def LUV_chroma_DE(LUV1, LUV2):
    du = np.power(LUV1[0] - LUV2[0], 2)
    dv = np.power(LUV1[1] - LUV2[1], 2)

    return np.sqrt(du + dv)


plank_lasso_uvY = np.zeros((len(plank_lasso), 2))
for c in range(len(plank_lasso)):
    uvY = xyY2uvY(np.array([plank_lasso[c, 0], plank_lasso[c, 1], 100.0]))
    plank_lasso_uvY[c, :] = np.array([uvY[0], uvY[1]])


import numpy as np
from sklearn.cluster import KMeans

import pycircstat as cstat
import astropy.stats.circstats as circstats

def max_sat(labs):
    """
    Calculates the maximum saturation of a given set of LAB colors.

    Parameters:
    labs (numpy.ndarray): A numpy array of shape (n, 3) containing n LAB colors.

    Returns:
    numpy.ndarray: A numpy array of shape (2,) containing the mean a* and b* values of the most saturated colors.
    """
    ls = labs[:, 0]
    As = labs[:, 1]
    bs = labs[:, 2]

    chromas = np.sqrt(np.power(As, 2) + np.power(bs, 2))
    sats = chromas/ls

    idxs = np.argsort(sats)
    five_perc_idxs = idxs[int(len(idxs)*0.95):]

    hs = np.array([bs[five_perc_idxs], As[five_perc_idxs]])
    hs[hs < 0] += 2*np.pi
    msat = cstat.mean(hs)

    return msat


def max_sat_hue(labs):
    """
    Calculates the maximum saturation hue of the input LAB color space image.

    Parameters:
    labs (numpy.ndarray): A numpy array of shape (n, 3) containing the LAB color space image.

    Returns:
    float: The maximum saturation hue of the input image.
    """
    ls = labs[:, 0]
    As = labs[:, 1]
    bs = labs[:, 2]

    chromas = np.sqrt(np.power(As, 2) + np.power(bs, 2))
    sats = chromas/ls

    idxs = np.argsort(sats)
    five_perc_idxs = idxs[int(len(idxs)*0.95):]

    hues = np.arctan2(bs[five_perc_idxs], As[five_perc_idxs])
    hues[hues < 0] += 2*np.pi

    msat = cstat.mean(hues)
    return msat


def most_lum(labs):
    """
    Calculates the most luminous color in the given LAB color space.

    Parameters:
    labs (numpy.ndarray): An array of LAB color values.

    Returns:
    numpy.ndarray: The most luminous color in the given LAB color space.
    """
    ls = labs[:, 0]
    As = labs[:, 1]
    bs = labs[:, 2]

    idxs = np.argsort(ls)
    five_perc_idxs = idxs[int(len(idxs)*0.95):]

    hs = np.array([bs[five_perc_idxs], As[five_perc_idxs]])
    hs[hs < 0] += 2*np.pi
    mlum = cstat.mean(hs)

    return mlum


def mean_col(labs):
    """
    Calculates the mean color of an image in LAB color space.

    Parameters:
    labs (numpy.ndarray): A numpy array of shape (n, 3) containing the LAB color values of an image.

    Returns:
    numpy.ndarray: A numpy array of shape (3,) containing the mean LAB color values of the image.
    """
    ls = labs[:, 0]
    As = labs[:, 1]
    bs = labs[:, 2]

    mcol = np.array([np.mean(ls), np.mean(As), np.mean(bs)])
    return mcol


def mean_hue(labs):
    """
    Calculates the mean hue of an array of LAB color values.

    Parameters:
    labs (numpy.ndarray): An array of LAB color values.

    Returns:
    float: The mean hue value in radians.
    """
    As = labs[:, 1]
    bs = labs[:, 2]

    hues = np.arctan2(bs, As)
    hues[hues < 0] += 2*np.pi

    return cstat.mean(hues)


def max_cov(labs):
    ls = labs[:, 0]
    As = labs[:, 1]
    bs = labs[:, 2]

    hues = np.arctan2(bs, As)
    hues[hues < 0] += 2*np.pi

    kmeans = KMeans(n_clusters=2, random_state=0, n_init="auto").fit(hues.reshape(-1, 1))
    cs = kmeans.labels_

    clust1 = np.where(cs == 0)[0]
    clust1_size = len(clust1)
    clust2 = np.where(cs == 1)[0]
    clust2_size = len(clust2)

    if clust1_size > clust2_size:
        lm = np.mean(ls[clust1])
        Am = np.mean(As[clust1])
        bm = np.mean(bs[clust1])

        return np.array([lm, Am, bm])
    else:
        lm = np.mean(ls[clust2])
        Am = np.mean(As[clust2])
        bm = np.mean(bs[clust2])

        return np.array([lm, Am, bm])


def max_cov_pixels(labs):
    ls = labs[:, 0]
    As = labs[:, 1]
    bs = labs[:, 2]

    hues = np.arctan2(bs, As)
    hues[hues < 0] += 2*np.pi

    kmeans = KMeans(n_clusters=2, random_state=0, n_init="auto").fit(hues.reshape(-1, 1))
    cs = kmeans.labels_

    clust1 = np.where(cs == 0)[0]
    clust1_size = len(clust1)
    clust2 = np.where(cs == 1)[0]
    clust2_size = len(clust2)

    if clust1_size > clust2_size:
        return np.array([ls[clust1], As[clust1], bs[clust1]]), cs
    else:
        return np.array([ls[clust2], As[clust2], bs[clust2]]), cs


def most_freq(labs):
    """
    Finds the most frequently occurring color in a given array of colors.

    Parameters:
    labs (numpy.ndarray): An array of colors in the format of (R, G, B).

    Returns:
    numpy.ndarray: An array representing the most frequently occurring color in the input array.
    """
    xs = labs[:, 0]
    ys = labs[:, 1]
    zs = labs[:, 2]

    xmin = np.min(xs) - 0.02
    ymin = np.min(ys) - 0.02
    zmin = np.min(zs) - 0.02

    xmax = np.max(xs) + 0.02
    ymax = np.max(ys) + 0.02
    zmax = np.max(zs) + 0.02

    x_sps = np.linspace(xmin, xmax, 10)
    y_sps = np.linspace(ymin, ymax, 10)
    z_sps = np.linspace(zmin, zmax, 10)

    step_size = 10

    max_freq = np.NINF
    xc_mfreq = 1
    yc_mfreq = 1
    zc_mfreq = 1

    for xc in range(len(x_sps)):
        x = x_sps[xc] + step_size/2
        for yc in range(len(y_sps)):
            y = y_sps[yc] + step_size/2
            for zc in range(len(z_sps)):
                z = z_sps[zc] + step_size/2
            
                cond1 = np.logical_and(xs > (x - step_size/2), xs < (x + step_size/2))
                cond2 = np.logical_and(ys > (y - step_size/2), ys < (y + step_size/2)) 
                cond3 = np.logical_and(zs > (z - step_size/2), zs < (z + step_size/2))
                idxs = np.logical_and(cond1, np.logical_and(cond2, cond3))
            
                ncolors = len(np.where(idxs)[0])
                if ncolors > max_freq:
                    max_freq = ncolors
                    xc_mfreq_coarse = xc
                    yc_mfreq_coarse = yc
                    zc_mfreq_coarse = zc

    xmin = xs[xc_mfreq_coarse] - step_size/2
    ymin = ys[yc_mfreq_coarse] - step_size/2
    zmin = zs[zc_mfreq_coarse] - step_size/2

    xmax = xs[xc_mfreq_coarse] + step_size/2
    ymax = ys[yc_mfreq_coarse] + step_size/2
    zmax = zs[zc_mfreq_coarse] + step_size/2

    x_sps = np.arange(xmin, xmax, 1)
    y_sps = np.arange(ymin, ymax, 1)
    z_sps = np.arange(zmin, zmax, 1)

    step_size = 1

    max_freq = np.NINF
    xc_mfreq = 1
    yc_mfreq = 1
    zc_mfreq = 1

    for xc in range(len(x_sps)):
        x = x_sps[xc] + step_size/2
        for yc in range(len(y_sps)):
            y = y_sps[yc] + step_size/2
            for zc in range(len(z_sps)):
                z = z_sps[zc] + step_size/2
            
                cond1 = np.logical_and(xs > (x - step_size/2), xs < (x + step_size/2))
                cond2 = np.logical_and(ys > (y - step_size/2), ys < (y + step_size/2))
                cond3 = np.logical_and(zs > (z - step_size/2), zs < (z + step_size/2))
                idxs = np.logical_and(cond1, np.logical_and(cond2, cond3))

                ncolors = len(np.where(idxs)[0])
                if ncolors > max_freq:
                    max_freq = ncolors
                    xc_mfreq = xc
                    yc_mfreq = yc
                    zc_mfreq = zc

    mfreq_col = [xs[xc_mfreq] + step_size/2, ys[yc_mfreq] + step_size/2, zs[zc_mfreq] + step_size/2]
    return np.array(mfreq_col)


# written for CIELAB
def classify_color(lab, color):
    l = lab[0]
    a = lab[1]
    b = lab[2]

    h = np.arctan2(b, a)
    if h < 0:
        h += 2 * np.pi

    hd = np.rad2deg(h)
    if color == 1:
        # 'Purple Center, Green Edge'
        if hd >= 80 and hd < 275:
            return "purple", "center"
        else:
            return "green", "edge"
    
    elif color == 2:
        # 'Pink Center, Orange Edge'
        # if a < 12 and b > 10:
        if hd >= 62:
            return "pink", "center"
        else:
            return "orange", "edge"

    elif color == 3:
        # 'Green Center, Purple Edge'
        if hd >= 80 and hd < 330:
            return "purple", "edge"
        else:
            return "green", "center"


def iri_extract_edge_center(labs, color):
    """
    Extracts the edge and center colors from an image in LAB color space.

    Args:
        labs (numpy.ndarray): A numpy array of shape (n, 3) containing the LAB color values of an image.
        color (str): The color to extract, either "red", "green", or "blue".

    Returns:
        dict: A dictionary containing the center and edge colors of the specified color, along with their indices.
            The dictionary has the following structure:
            {
                'cent': {
                    'lab': numpy.ndarray,
                    'idx': list,
                },
                'edge': {
                    'lab': numpy.ndarray,
                    'idx': list,
                }
            }
    """
    ls = labs[:, 0]
    As = labs[:, 1]
    bs = labs[:, 2]

    edge_center = {
        'cent': {
            'lab': [],
            'idx': [],
        },
        'edge': {
            'lab': [],
            'idx': [],
        }
    }

    for c in range(len(labs)):
        cat, cent_edge = classify_color(labs[c, :].squeeze(), color)
        if cent_edge == "center":
            edge_center['cent']['lab'].append([ls[c], As[c], bs[c]])
            edge_center['cent']['idx'].append(c)
        else:
            edge_center['edge']['lab'].append([ls[c], As[c], bs[c]])
            edge_center['edge']['idx'].append(c)

    edge_center['cent']['lab'] = np.array(edge_center['cent']['lab'])
    edge_center['edge']['lab'] = np.array(edge_center['edge']['lab'])

    return edge_center


def yo_cent_edge_colors(img, mask, color):
    cent_yx = np.array([img.shape[0] / 2, img.shape[1] / 2])

    labs = np.zeros(img.shape)
    for y in range(img.shape[0]):
        for x in range(img.shape[1]):
            if mask[y, x]:
                labs[y, x, :] = robsblobs.infamous_lab.rgb2lab(mon, img[y, x, :])

    cent_map = np.zeros(img.shape[:2], dtype=bool)
    edge_map = np.zeros(img.shape[:2], dtype=bool)

    h2 = img.shape[0]/2
    w2 = img.shape[1]/2
    for theta in np.linspace(0, 2*np.pi, 1000):
        for r in np.arange(0, h2, 0.5):
            y = r*np.sin(theta) + (h2-1)
            x = r*np.cos(theta) + (w2-1)
            if mask[int(y), int(x)]:
                cat, cent_edge = robsblobs.img_stats.classify_color(labs[int(y), int(x), :], color)
                if cent_edge == "center":
                    cent_map[int(y), int(x)] = True
                else:
                    edge_map[int(y), int(x)] = True

    return cent_map, edge_map


import numpy as np

def X(x, y, Y):
    return (x/y) * Y


def Z(x, y, Y):
    return ((1 - x - y)/y) * Y


def xyY2XYZ(xyY):
    return [X(xyY[0], xyY[1], xyY[2]), xyY[2], Z(xyY[0], xyY[1], xyY[2])]


def p1(y, s1):
    return v1(s1) - v1(y)


def p2(y, vy, s3):
    if s3 <= y:
        return v1(y) - v1(s3)
    else:
        return v2(y, vy) - v2(s3, vy)
    

def v1(x):
    if x >= 0.0043:
        return np.power(100 * x - 0.43, 0.51) / (np.power(100 * x - 0.43, 0.51) + 31.75)
    else:
        return 0


def v2(x, vy):
    if x > 0.001 * k(vy):
        return np.power(100 * x/k(vy) - 0.1, 0.86) / (np.power(100 * x/k(vy) - 0.1, 0.86) + 103.2)
    else:
        return 0


def k(vy):
    return 0.140 + 0.175*vy


def f1(y, vy, s1, s3):
    return 700*p1(y, s1) - 54*p2(y, vy, s3)


def f2(y, vy, s3):
    return 96.5*p2(y, vy, s3)


def svf(txyz, bxyz):
    y = txyz[1]

    xyz2lms = np.array([[ 0.520,  0.589, -0.102],
                        [-0.194,  0.562,  0.034],
                        [ 0.007, -0.015,  0.907]])

    bxyz = bxyz/bxyz[1]

    tlms = xyz2lms @ txyz
    blms = xyz2lms @ bxyz
    tlmsn = tlms/blms

    s1 = tlmsn[0]
    s2 = tlmsn[1]
    s3 = tlmsn[2]

    vy = 40*v1(y)
    
    return np.array([f1(y, vy, s1, s3), f2(y, vy, s3), vy])


def svfde(tc1, bc1, tc2, bc2):
    svf1 = svf(tc1, bc1)
    c1f1 = svf1[0]
    c1f2 = svf1[1]
    c1vy = svf1[2]

    svf2 = svf(tc2, bc2)
    c2f1 = svf2[0]
    c2f2 = svf2[1]
    c2vy = svf2[2]

    return np.sqrt(np.power(c1f1 - c2f1, 2) + np.power(c1f2 - c2f2, 2) + np.power(2.3*(c1vy - c2vy), 2))


def svfsat(f1, f2, vy):
    return np.sqrt(np.power(f1, 2) + np.power(f2, 2))/vy


import numpy as np

from .monitor import Monitor
from .cie_standard import xyY2XYZ

def rgb2xyz(mon: Monitor, rgb):
    """
    Convert RGB color space to XYZ color space using the given monitor's RGB2XYZ matrix.

    Args:
        mon (Monitor): The monitor object containing the RGB2XYZ matrix.
        rgb (tuple): The RGB color to be converted.

    Returns:
        tuple: The converted XYZ color.
    """
    return mon.RGB2XYZ @ rgb


def xyz2rgb(mon: Monitor, xyz):
    """
    Convert a color from CIE XYZ color space to RGB color space.

    Args:
        mon (Monitor): The monitor object containing the color space conversion matrix.
        xyz (tuple): A tuple containing the CIE XYZ color values.

    Returns:
        tuple: A tuple containing the RGB color values.
    """
    return mon.XYZ2RGB @ xyz


def xyY2rgb(mon: Monitor, xyY):
    """
    Convert xyY color space to RGB color space.

    Args:
        mon (Monitor): The monitor object.
        xyY (tuple): A tuple containing the x, y, and Y values.

    Returns:
        tuple: A tuple containing the R, G, and B values.
    """
    return xyz2rgb(mon, xyY2XYZ(xyY))


def luminance(mon: Monitor, rgb):
    """
    Calculates the luminance of a given RGB color on a given monitor.

    Args:
        mon (Monitor): The monitor object representing the monitor on which the color is displayed.
        rgb (tuple): A tuple representing the RGB color to calculate the luminance for.

    Returns:
        float: The calculated luminance value.
    """
    lums = rgb * [mon.monxyY[0, 2], mon.monxyY[1, 2], mon.monxyY[2, 2]]
    return np.array(lums).sum()/mon.maxLuminance



import numpy as np

# helper functions for computing DKL2RGB in monitor.py
# from Qasim Zaidi
def lumchrm(lumr, lumg, lumb, r, g, b):
  lum = np.array([lumr, lumg, lumb])

  bigl = 0.0
  bigm = 0.0
  bigs = 0.0

  for i in range(3):
     print('Luminances[', i, '] = ', lum[i])
     lu = lum[i]
     bigl = bigl + lu*r[i]
     bigm = bigm + lu*g[i]
     bigs = bigs + lu*b[i]

  denom = bigl+bigm
  print('bigl =', bigl ,' bigm =', bigm, ' bigs =', bigs)

  lout = bigl/denom
  sout = bigs/denom
  print('L/L+M =', lout, ' S/L+M =', sout)

  return


def solvex(a, b, c, d, e, f):
  return (a*f/d-b)/(c*f/d-e)


def solvey(a, b, c, d, e, f):
  return (a*e/c-b)/(d*e/c-f)
  

def cie2lms(x, y):
  """
  Convert CIE color space coordinates to LMS color space coordinates.

  Args:
    x (float): The x coordinate in CIE color space.
    y (float): The y coordinate in CIE color space.

  Returns:
    numpy.ndarray: A 1D array of length 3 containing the LMS color space coordinates.
  """
  z = 1-x-y
  cie = np.array([x, y, z]).T

  matrix = np.array([[ .15514, .54316, -.03286],
             [-.15514, .45684,  .03286],
             [  0,     0,       .01608]])

  lms = matrix @ cie

  l = lms[0]/(lms[0]+lms[1]) #   L/(L+M)
  m = lms[1]/(lms[0]+lms[1]) #   M/(L+M)
  s = lms[2]/(lms[0]+lms[1]) #   S/(L+M)
  norm = np.array([l, m, s])

  return norm


import numpy as np
from scipy.interpolate import CubicSpline

from .cmfs import ciexyz_1931

konica_wlns = np.linspace(380, 781)

wavelength_cmf = np.array(list(ciexyz_1931[0]))
x_bar = np.array(list(ciexyz_1931[1]))
y_bar = np.array(list(ciexyz_1931[2]))
z_bar = np.array(list(ciexyz_1931[3]))

x_interp = CubicSpline(wavelength_cmf, x_bar)
x_new_cmf = x_interp(konica_wlns)
x_new_cmf[x_new_cmf < 0] = 0

y_interp = CubicSpline(wavelength_cmf, y_bar)
y_new_cmf = y_interp(konica_wlns)
y_new_cmf[y_new_cmf < 0] = 0

z_interp = CubicSpline(wavelength_cmf, z_bar)
z_new_cmf = z_interp(konica_wlns)
z_new_cmf[z_new_cmf < 0] = 0

diffs = np.diff(konica_wlns)
stepSize = np.zeros((len(diffs)+1))
stepSize[:len(diffs)] = diffs

# assume always using konica wln range for now
def spect2xyz(spect):
    """
    Convert spectral data to CIE XYZ color space.

    Parameters:
    spect (numpy.ndarray): Spectral data to be converted.

    Returns:
    numpy.ndarray: The converted CIE XYZ color space data.
    """
    corrCMF = np.zeros((len(x_new_cmf), 3))
    corrCMF[:, 0] = 683 * x_new_cmf * stepSize
    corrCMF[:, 1] = 683 * y_new_cmf * stepSize
    corrCMF[:, 2] = 683 * z_new_cmf * stepSize

    return spect.T @ corrCMF


import numpy as np

from .monitor import Monitor

def matlab_adapt_xyz(mon: Monitor, xyz_in):
    wp_D65_matlab = np.array([0.9505, 1.0000, 1.0888])

    # Bradford cone response model matrix
    Ma = np.array([
        [0.8951, 0.2664, -0.1614],
        [-0.7502, 1.7135, 0.0367],
        [0.0389, -0.0685, 1.0296],
    ])

    # Source white point cone response
    CRs = Ma @ wp_D65_matlab.T
    
    # Destination white point cone response
    CRd = Ma @ mon.monWP.T
    
    # Cone response domain scaling matrix
    S = np.diag(CRd / CRs)

    # Linear adaptation transform matrix
    M = np.linalg.lstsq(Ma, S @ Ma)

    return xyz_in @ M[0].T


def matlab_rgb2xyz(mon, rgb):
    rgb2xyz_mat = np.array([
        [0.4125, 0.3576, 0.1804],
        [0.2127, 0.7152, 0.0722],
        [0.0193, 0.1192, 0.9503],
    ])

    xyz_pre = rgb @ rgb2xyz_mat.T
    return matlab_adapt_xyz(mon, xyz_pre)



import numpy as np

def rgb2hsv(rgb):
    """
    Convert an RGB color value to HSV.

    Args:
        rgb (tuple): A tuple containing the red, green, and blue values (in that order) of the color to be converted.

    Returns:
        numpy.ndarray: A 1D numpy array containing the hue, saturation, and value (in that order) of the converted color.
    """
    R = rgb[0]
    G = rgb[1]
    B = rgb[2]

    M = np.max(rgb)
    m = np.min(rgb)
    C = M - m

    S = 0
    Hprime = 0
    if C == 0:
        return np.array([Hprime, S, M])
    elif M == R:
        Hprime = np.mod((G - B)/C, 6)
    elif M == G:
        Hprime = (B - R)/C + 2
    elif M == B:
        Hprime = (R - G)/C + 4
    
    H = 60 * Hprime
    
    alpha = R - G * np.cos(np.deg2rad(60)) - B * np.cos(np.deg2rad(60))
    beta = G * np.sin(np.deg2rad(60)) - B * np.sin(np.deg2rad(60))
    H2 = np.arctan2(beta, alpha)
    if H2 < 0:
        H2 += 2*np.pi

    C2 = np.sqrt(np.power(alpha, 2) + np.power(beta, 2))

    V = M

    if V != 0:
        # S = C2/V
        S = C/V

    return np.array([H, S, V])


import numpy as np
from scipy.interpolate import CubicSpline, LinearNDInterpolator

from .spectra import spect2xyz
from .illuminants import d65_xie_macadam
from .cie_standard import XYZ2xyY

minw = 380
maxw = 781
nwlns = 1000
wlns = np.linspace(minw, maxw, nwlns)
ncents = 100
nwidths = 50

cents = np.linspace(minw, maxw, ncents)
widths = np.linspace(1, maxw-minw, nwidths)

spect = np.zeros((ncents, nwidths, 2, len(wlns)))
for i in range(ncents):
    for j in range(nwidths):
        s = np.select(
            [
                np.logical_and(wlns >= cents[i]-widths[j]/2, wlns <= cents[i]+widths[j]/2), 
                np.logical_and(wlns <= cents[i]-widths[j]/2, wlns >= cents[i]+widths[j]/2)
            ],
            [
                1,
                0
            ]
        )

        spect[i, j, 0, :] = s
        spect[i, j, 1, :] = 1-s


konica_wlns = np.linspace(380, 781)
xyz = np.zeros((ncents, nwidths, 2, 3))
xyY = np.zeros((ncents, nwidths, 2, 3))
for i in range(ncents):
    for j in range(nwidths):
        s_interp = CubicSpline(wlns, spect[i, j, 0, :])
        s_new = s_interp(konica_wlns)
        xyz[i, j, 0, :] = spect2xyz(s_new * d65_xie_macadam)
        xyY[i, j, 0, :] = XYZ2xyY(xyz[i, j, 0, :])

        s_interp = CubicSpline(wlns, spect[i, j, 1, :])
        s_new = s_interp(konica_wlns)
        xyz[i, j, 1, :] = spect2xyz(s_new * d65_xie_macadam)
        xyY[i, j, 1, :] = XYZ2xyY(xyz[i, j, 1, :])


xs = xyY[:, :, :, 0].flatten()
ys = xyY[:, :, :, 1].flatten()
zs = xyY[:, :, :, 2].flatten()
macadam_interp = LinearNDInterpolator(list(zip(xs, ys)), zs)


import numpy as np

from .monitor import Monitor
from .infamous_lab import rgb2lab
from .lch import lab2lch

# https://stackoverflow.com/questions/596216/formula-to-determine-perceived-brightness-of-rgb-color
# /**
#  * Saturated colors appear brighter to human eye.
#  * That's called Helmholtz-Kohlrausch effect.
#  * Fairchild and Pirrotta came up with a formula to
#  * calculate a correction for that effect.
#  * "Color Quality of Semiconductor and Conventional Light Sources":
#  * https://books.google.ru/books?id=ptDJDQAAQBAJ&pg=PA45&lpg=PA45&dq=fairchild+pirrotta+correction&source=bl&ots=7gXR2MGJs7&sig=ACfU3U3uIHo0ZUdZB_Cz9F9NldKzBix0oQ&hl=ru&sa=X&ved=2ahUKEwi47LGivOvmAhUHEpoKHU_ICkIQ6AEwAXoECAkQAQ#v=onepage&q=fairchild%20pirrotta%20correction&f=false
#  * @return {number}
#  */
# function getLightnessUsingFairchildPirrottaCorrection([l, c, h]) {
#     const l_ = 2.5 - 0.025 * l
#     const g = 0.116 * Math.abs(Math.sin(degreesToRadians((h - 90) / 2))) + 0.085
#     return l + l_ * g * c
# }

# export function getPerceivedLightness(hex) {
#     return getLightnessUsingFairchildPirrottaCorrection(labToLch(xyzToLab(rgbToXyz(hex))))
# }

# Converts RGB values (between 0 and 1) first to LCh, then to the Fairchild L** measure for perceived brightness
def rgb2FairchildLstar(mon: Monitor, rgb):
    lab = rgb2lab(mon, rgb)
    l = lab[0]
    
    lch = lab2lch(lab)
    chrom = lch[1]
    H = lch[2]
    
    Hd = np.rad2deg(H)
    H_corr = np.deg2rad((Hd - 90)/2)
        
    # calculate the two functions: 
    # f1 provides the hue dependency, which looks like the absolute value of a
    # sin wave (half a cycle), i.e. the effect drops down at 90 degrees, i.e.
    # at yellow

    f1 = 0.116 * np.abs(np.sin(H_corr)) + 0.085

    # f2 allows the prediction of the HK effect to be much larger at low
    # luminance factors than at high luminance factors

    f2 = 2.5 - 0.025*l

    return l + (f2 * f1 * chrom)


from . import infamous_lab
from . import cie_standard
from . import cie_monitor_helpers
from . import fairchildL
from . import funny_dude_discovers_maxRGB_in_adobe
from . import G0_brilliance
from . import ware_cowan
from . import lms
from . import maxRGB_revolution
from . import sumrgb
from . import nayatani_insanity
from . import radiance
from . import uv_space
from . import cmfs
from . import plotting
from . import brightness_ranking
from . import lch
from . import guth_vector
from . import jazzy
from . import hunt
from . import spectra
from . import illuminants
from . import hsv
from . import jingHSP
from . import svf
from . import hue_angles
from . import matlab
from . import img_stats
from . import atd95
from . import strobe_export
from . import monitor_fitting
from . import alices_damn_monitor

# from . import nayatani_test
# from . import hunt_test
# from . import atd95_test

cie_standard.generatePlankLasso()


import numpy as np

sRGB_xyY = np.array([[0.6400, 0.3300, 0.2126*100],
                    [0.3000, 0.6000, 0.7152*100],
                    [0.1500, 0.0600, 0.0722*100]])

sRGB_WP = np.array([0.3127, 0.3290, 1.0000*100])

sRGB2XYZ = np.array([[0.4124, 0.3576, 0.1805],
                     [0.2126, 0.7152, 0.0722],
                     [0.0193, 0.1192, 0.9505]])

XYZ2sRGB = np.linalg.inv(sRGB2XYZ)

def xyz2srgb_linear(XYZ):
    """
    Converts linear XYZ values to linear sRGB values using the XYZ2sRGB matrix.

    Args:
        XYZ (numpy.ndarray): A 3-element array containing the linear XYZ values.

    Returns:
        numpy.ndarray: A 3-element array containing the linear sRGB values.
    """
    print("xyz2srgb_linear: linear srgb values!")
    return XYZ2sRGB @ XYZ


def linear_srgb2XYZ(srgb):
    """
    Convert linear sRGB values to XYZ color space using the sRGB2XYZ matrix.

    Args:
        srgb (numpy.ndarray): Linear sRGB values as a 3-element array.

    Returns:
        numpy.ndarray: XYZ color space values as a 3-element array.
    """
    print("linear_srgb2XYZ: needs linear srgb values!")
    return sRGB2XYZ @ srgb


def srgb_GammaCorrect(srgb):
    """
    Applies gamma correction to an sRGB color.

    Args:
        srgb (numpy.ndarray): An array of sRGB color values.

    Returns:
        numpy.ndarray: An array of gamma-corrected sRGB color values.
    """
    srgb_gc = np.zeros((3))
    cc = 0
    for c in srgb:
        if c <= 0.0031308:
            srgb_gc[cc] = 12.92*c
        else:
            srgb_gc[cc] = 1.055 * np.power(c, 1/2.4) - 0.055

        cc += 1
    
    return srgb_gc


def srgb_linearize(srgb_gc):
    """
    Linearize sRGB gamma-compressed values.

    Args:
        srgb_gc (numpy.ndarray): Array of sRGB gamma-compressed values.

    Returns:
        numpy.ndarray: Array of linearized sRGB values.
    """
    srgb = np.zeros((3))
    cc = 0
    for c in srgb_gc:
        if c <= 0.04045:
            srgb[cc] = c/12.92
        else:
            srgb[cc] = np.power((c + 0.055)/1.055, 2.4)

        cc += 1

    return srgb


XYZ2CIERGB = np.array([[8041697, -3049000, -1591845],
                       [-1752003, 4851000, 301853],
                       [17697, -49000, 3432153]])
XYZ2CIERGB = (1/3400850) * XYZ2CIERGB

CIERGB2XYZ = np.linalg.inv(XYZ2CIERGB)

def xyz2ciergb(XYZ):
    """
    Convert XYZ color space to CIE RGB color space.

    Parameters:
    XYZ (numpy.ndarray): An array of shape (3,) representing the XYZ color space.

    Returns:
    numpy.ndarray: An array of shape (3,) representing the CIE RGB color space.
    """
    return XYZ2CIERGB @ XYZ


def ciergb2xyz(rgb):
    """
    Convert a color from CIE RGB color space to CIE XYZ color space.

    Args:
        rgb (tuple): A tuple of three values representing the red, green, and blue components of the color.

    Returns:
        tuple: A tuple of three values representing the X, Y, and Z components of the color in the CIE XYZ color space.
    """
    return CIERGB2XYZ @ rgb


import numpy as np

# found this by accident:
# https://alienryderflex.com/hsp.html
# i wonder what other surprises lie out there

Pr = 0.241
Pg = 0.691
Pb = 0.068

Pr_informed = 0.299
Pg_informed = 0.587
Pb_informed = 0.114

Pr_jing = 0.302
Pg_jing = 0.550
Pb_jing = 0.148

Pr_newJing = 0.2667
Pg_newJing = 0.6267
Pb_newJing = 0.1067

# public domain function by Darel Rex Finley, 2006
#
# This function expects the passed-in values to be on a scale
# of 0 to 1, and uses that same scale for the return values.
#
# See description/examples at alienryderflex.com/hsp.html
def RGBtoHSP(R, G, B, version):
  # Calculate the Perceived brightness.
  if version == "original":
    P = np.sqrt(R*R*Pr + G*G*Pg + B*B*Pb)
  elif version == "photoshop":
    P = np.sqrt(R*R*Pr_informed + G*G*Pg_informed + B*B*Pb_informed)
  elif version == "jing":
    P = np.sqrt(R*R*Pr_jing + G*G*Pg_jing + B*B*Pb_jing)
  elif version == "jing_new":
    P = np.sqrt(R*R*Pr_newJing + G*G*Pg_newJing + B*B*Pb_newJing)

  # Calculate the Hue and Saturation. (This part works
  # the same way as in the HSV/B and HSL systems???.)
  if R==G and R==B:
    H = 0.0
    S = 0.0
    return np.array([H, S, P])

  if R>=G and R>=B:              # R is largest
    if B>=G:
      H = 6/6 - 1/6*(B-G)/(R-G)
      S = 1 - G/R
    else:
      H = 0/6 + 1/6*(G-B)/(R-B)
      S = 1 - B/R
  elif G >= R and G >= B:        # G is largest
    if R>=B:
      H = 2/6 - 1/6*(R-B)/(G-B)
      S = 1 - B/G
    else:
      H = 2/6 + 1/6*(B-R)/(G-R)
      S = 1 - R/G
  else:                          # B is largest
    if G>=R:
      H = 4/6 - 1/6*(G-R)/(B-R)
      S = 1 - R/B
    else:
      H = 4/6 + 1/6*(R-G)/(B-G)
      S = 1 - G/B

  return np.array([H, S, P])


# public domain function by Darel Rex Finley, 2006
#
# This function expects the passed-in values to be on a scale
# of 0 to 1, and uses that same scale for the return values.
#
# Note that some combinations of HSP, even if in the scale
# 0-1, may return RGB values that exceed a value of 1.  For
# example, if you pass in the HSP color 0,1,1, the result
# will be the RGB color 2.037,0,0.
#
# See description/examples at alienryderflex.com/hsp.html
def HSPtoRGB(H, S, P):
  part = 0
  minOverMax = 1 - S

  if minOverMax>0.:
    if H < 1/6:                                                  # R>G>B
      H = 6*(H - 0/6)
      part = 1 + H*(1/minOverMax - 1)
      B = P/np.sqrt(Pr/minOverMax/minOverMax + Pg*part*part + Pb)
      R = B/minOverMax
      G = B + H*(R - B)
    elif H < 2/6:                                                # G>R>B
      H = 6*(-H + 2/6)
      part = 1 + H*(1/minOverMax - 1)
      B = P/np.sqrt(Pg/minOverMax/minOverMax + Pr*part*part + Pb)
      G = B/minOverMax
      R = B + H*(G - B)
    elif H < 3/6:                                                # G>B>R
      H = 6*(H - 2/6)
      part = 1 + H*(1/minOverMax - 1)
      R = P/np.sqrt(Pg/minOverMax/minOverMax + Pb*part*part + Pr)
      G = R/minOverMax
      B = R + H*(G - R)
    elif H < 4/6:                                                # B>G>R
      H = 6*(-H + 4/6)
      part = 1 + H*(1/minOverMax-1)
      R = P/np.sqrt(Pb/minOverMax/minOverMax + Pg*part*part + Pr)
      B = R/minOverMax
      G = R+H*(B-R)
    elif H < 5./6.:                                              # B>R>G
      H = 6*(H - 4/6)
      part = 1 + H*(1/minOverMax-1)
      G = P/np.sqrt(Pb/minOverMax/minOverMax + Pr*part*part + Pg)
      B = G/minOverMax
      R = G+H*(B-G)
    else:                                                        # R>B>G
      H = 6*(-H + 6/6)
      part = 1 + H*(1/minOverMax-1)
      G = P/np.sqrt(Pr/minOverMax/minOverMax + Pb*part*part + Pg)
      R = G/minOverMax
      B = G+H*(R-G)
  else:
    if H < 1/6:                              # R>G>B
      H = 6*(H - 0/6)
      R = np.sqrt(P*P/(Pr + Pg*H*H))
      G = R*H
      B = 0
    elif H < 2/6:                            # G>R>B
      H = 6*(-H + 2/6)
      G = np.sqrt(P*P/(Pg + Pr*H*H))
      R = G*H
      B = 0
    elif H < 3/6:                            # G>B>R
      H = 6*(H - 2/6)
      G = np.sqrt(P*P/(Pg + Pb*H*H))
      B = G*H
      R = 0
    elif H < 4/6:                            # B>G>R
      H = 6*(-H + 4/6)
      B = np.sqrt(P*P/(Pb + Pg*H*H))
      G = B*H
      R = 0
    elif H < 5/6:                            # B>R>G
      H = 6*(H - 4/6)
      B = np.sqrt(P*P/(Pb + Pr*H*H))
      R = B*H
      G = 0
    else:                                    # R>B>G
      H = 6*(-H + 6/6)
      R = np.sqrt(P*P/(Pr + Pb*H*H))
      B = R*H
      G = 0

  return np.array([R, G, B])


import numpy as np
import pandas as pd
import glob
import scipy.io as sio

from .monitor import Monitor
from .cie_monitor_helpers import rgb2xyz

flickerTypes = dict([
    (0, "RedGreen"),
    (1, "RedPurple"),
    (2, "GreenPurple"),
    (3, "GrayGreen"),
    (4, "GrayPurple"),
    (5, "GrayYellow"),
    (6, "GreenYellow"),
    (7, "GrayRed"),
    (8, "GrayBlue"),
    (9, "GrayCyan")
])

motionTypes = dict([
    (0, "RedGreen"),
    (1, "GreenYellow"),
    (2, "PurpleGreen"),
    (3, "GreenCyan")
])

def convert_color_weight_to_lum(mon, df):
    w = df['Weight']
    c = df['Color'][0]

    rgb_gc = np.zeros(3)
    rgb = np.zeros(3)
    for rc in range(3):
        # because this is what we sent to the monitor
        rgb_gc[rc] = np.power(w * c[rc], 1.0/2.2)
        rgb[rc] = np.power(rgb_gc[rc], mon.monGamma[rc])
            
    xyz = rgb2xyz(mon, rgb)
    return xyz[1]


def parse_flicker_data():
    pass
    #         for c in range(2):
    #             t = pd.DataFrame({
    #                 "StimType": "Flicker",
    #                 "SubjectName": str(db.iloc[tc]["SubjectName"]),
    #                 "BlockNumber": db.iloc[tc]["BlockNumber"],
    #                 "TrialInBlock": db.iloc[tc]["TrialInBlock"],
    #                 "TotalTrials": db.iloc[tc]["TotalTrials"],
    #                 "ReactionTime": float(db.iloc[tc]["ReactionTime"]),
    #                 "Flick": "First" if c == 0 else "Second",
    #                 "Weight": float(resp['weight_first']) if c == 0 else float(resp['weight_second']),
    #                 "Color": [resp['first_col'] if c == 0 else resp['second_col']],
    #                 "FlickerType": flickerTypes[resp['flicker_type']]
    #             })
    #             df = pd.concat([df, t])


def parse_minmotion_data(mon: Monitor, data, resp, tc):
    t = []
    for c in range(2):
        df = pd.DataFrame({
            "StimType": "MinimalMotion",
            "SubjectName": data["SubjectName"][tc],
            "BlockNumber": data["BlockNumber"][tc],
            "TrialInBlock": data["TrialInBlock"][tc],
            "TotalTrials": data["TotalTrials"][tc],
            "ReactionTime": data["ReactionTime"][tc],
            "ColorComponent": "First" if c == 0 else "Second",
            "Weight": float(resp['weight_first'] if c == 0 else resp['weight_second']),
            "Color": [resp['first_col'] if c == 0 else resp['second_col']],
            "MotionType": motionTypes[resp['motion_type']]
        })

        df["Luminances"] = convert_color_weight_to_lum(mon, df)

        t.append(df)

    return t
    

def parse_unique_yellow_data():
    pass
    # t = pd.DataFrame({
    #         "StimType": "UniqueYellow",
    #         "SubjectName": str(db.iloc[tc]["SubjectName"]),
    #         "BlockNumber": db.iloc[tc]["BlockNumber"],
    #         "TrialInBlock": db.iloc[tc]["TrialInBlock"],
    #         "TotalTrials": db.iloc[tc]["TotalTrials"],
    #         "ReactionTime": float(db.iloc[tc]["ReactionTime"]),
    #         "WeightR": [float(resp["weight_R"])],
    #         "WeightG": [float(resp["weight_G"])]
    #     })
    #     df = pd.concat([df, t])


def parse_brightness_sorting_data(mon: Monitor, data, resp, tc):
    indexes = []
    rs = []
    gs = []
    bs = []
    for rc in range(len(resp)):
        indexes.append(resp[rc]["rect_index"])
        rs.append(float(resp[rc]["color"][0]))
        gs.append(float(resp[rc]["color"][1]))
        bs.append(float(resp[rc]["color"][2]))

    t = pd.DataFrame({
        "StimType": "BrightnessSorting",
        "SubjectName": str(data.iloc[tc]["SubjectName"]),
        "BlockNumber": data.iloc[tc]["BlockNumber"],
        "TrialInBlock": data.iloc[tc]["TrialInBlock"],
        "TotalTrials": data.iloc[tc]["TotalTrials"],
        "ReactionTime": float(data.iloc[tc]["ReactionTime"]),
        "RectangleIndex": [indexes],
        "R": [rs],
        "G": [gs],
        "B": [bs],
    })
    
    return t


def parse_strobe_data(mon: Monitor, fn):
    d = pd.read_json(fn)

    d = d.rename({
        0: "SubjectName",
        1: "BlockNumber",
        2: "TrialInBlock",
        3: "TotalTrials",
        4: "ObserverResponse",
        5: "ReactionTime",
        6: "DontKnow",
    }, axis=1)
    d = d.drop(columns=["DontKnow"])

    sname = str(d["SubjectName"][0])

    df = pd.DataFrame()
    for tc in d["TotalTrials"]:
        resp = d.iloc[tc]["ObserverResponse"]
        if len(resp) == 0:
            continue

        ts = []
        if 'motion_type' in resp:
            ts = parse_minmotion_data(mon, d, resp, tc)
        elif 'weight_R' in resp:
            continue
            ts = parse_unique_yellow_data(mon, d, resp, tc)
        elif 'rect_index' in resp[0]:
            ts = parse_brightness_sorting_data(mon, d, resp, tc)
        elif 'flicker_type' in resp:
            continue
            ts = parse_flicker_data(mon, d, resp, tc)

        if isinstance(ts, list) and len(ts) > 0:
            for t in ts:
                df = pd.concat([df, t])
        else:
            df = pd.concat([df, ts])
        
    df = df.set_axis([str(k) for k in range(len(df))], axis='index')

    return df, sname


def export_strobe_data_to_json(fn, df):
    out = df.to_json(orient='records')
    with open(fn, "w") as f:
        f.write(out)


def export_strobe_data_to_matlab(fn, df):
    sio.savemat(fn, {'data': df.to_dict("list")})


import numpy as np

from .cmfs import ciexyz_1931

def XYZ2xyY(XYZ):
    """
    Convert CIE XYZ color space to CIE xyY color space.

    Parameters:
    XYZ (numpy.ndarray): An array of CIE XYZ values.

    Returns:
    numpy.ndarray: An array of CIE xyY values.
    """
    X = XYZ[0]
    Y = XYZ[1]
    Z = XYZ[2]

    s = X + Y + Z

    x = X/s
    y = Y/s

    return np.array([x, y, Y])


def XYZ2xy(XYZ):
    """
    Converts CIE 1931 XYZ tristimulus values to chromaticity coordinates (x, y).

    Parameters:
    XYZ (numpy.ndarray): Array of CIE 1931 XYZ tristimulus values.

    Returns:
    numpy.ndarray: Array of chromaticity coordinates (x, y).
    """
    xyY = XYZ2xyY(XYZ)
    x = xyY[0]
    y = xyY[1]

    return np.array([x, y])


def xyY2XYZ(xyY):
    """
    Convert xyY color space to XYZ color space.

    Args:
        xyY (numpy.ndarray): Array of xyY values.

    Returns:
        numpy.ndarray: Array of XYZ values.
    """
    x = xyY[0]
    y = xyY[1]
    Y = xyY[2]

    X = (Y/y) * x
    Z = (Y/y) * (1 - x - y)

    return np.array([X, Y, Z])


# generate the Planck-ian Locus (lasso) with the old school XYZ CMFs
plank_lasso = np.zeros((len(ciexyz_1931)+4, 2))
plank_colors = np.zeros((len(plank_lasso), 3))
def generatePlankLasso():
    """
    Generates a line of colors representing the Planckian locus in the CIE 1931 color space.

    Returns:
    None
    """
    for c in range(len(ciexyz_1931)):
        spect = np.zeros((len(ciexyz_1931), ))
        spect[c] = 1.0

        X = (np.array(list(ciexyz_1931[1])) * spect).sum()
        Y = (np.array(list(ciexyz_1931[2])) * spect).sum()
        Z = (np.array(list(ciexyz_1931[3])) * spect).sum()

        x = X/(X + Y + Z)
        y = Y/(X + Y + Z)

        plank_lasso[c, 0] = x
        plank_lasso[c, 1] = y

    # linearly interpolate between the first and last point to create the line of purples
    c = len(ciexyz_1931)
    for a in np.linspace(0, 1, 4):
        b = a*plank_lasso[0, :] + (1-a)*plank_lasso[len(ciexyz_1931)-1, :]
        plank_lasso[c, :] = b

        c += 1

    # r = 0
    # for c in range(len(plank_lasso)):
    #     xyY = np.array([plank_lasso[c, 0], plank_lasso[c, 1], 100])
    #     XYZ = xyY2XYZ(xyY)
    #     xyz2lab(mon: Monitor, XYZ)
    #     rgb = xyz2srgb(XYZ)
    #     plank_colors[c, :] = rgb


import numpy as np
import pandas as pd
from scipy.interpolate import CubicSpline

import importlib.resources

datadir = importlib.resources.files('robsblobs').joinpath('data')

# ciexyz_1931 = pd.read_csv('./ciexyz31_1.csv', header=None)
ciexyz_1931 = pd.read_csv(datadir.joinpath('ciexyz31_1.csv'), header=None)
# xyz_jv = pd.read_csv('./ciexyzjv.csv', header=None)
xyz_jv = pd.read_csv(datadir.joinpath('ciexyz31_1.csv'), header=None)

# lms_absorp = np.genfromtxt('linss2_10e_1.csv', delimiter=',')
lms_absorp = np.genfromtxt(datadir.joinpath('ciexyz31_1.csv'), delimiter=',')
lms_wlns = lms_absorp[0:391, 0]
l_absorp = lms_absorp[0:391, 1]
m_absorp = lms_absorp[0:391, 2]
s_absorp = lms_absorp[0:391, 3]

# hacky way to get vlambda for now
konica_wlns = np.arange(380, 781)
wlns = np.array(list(ciexyz_1931[0]))
Y = np.array(list(ciexyz_1931[2]))
Y_interp = CubicSpline(wlns, Y)
V_lambda = Y_interp(konica_wlns)


import numpy as np

from .monitor import Monitor
from .cie_monitor_helpers import rgb2xyz, xyz2rgb
from .matlab import matlab_rgb2xyz

lab_delta = 6/29
def lab_f(t):
    if t > np.power(lab_delta, 3):
        return np.power(t, 0.3333333333333333)
    else:
        return t/(3 * (np.power(lab_delta, 2))) + 4/29


def xyz2lab(mon: Monitor, XYZ):
    xyz_s = XYZ / mon.monWP
    xsf = lab_f(xyz_s[0])
    ysf = lab_f(xyz_s[1])
    zsf = lab_f(xyz_s[2])

    l = 116 * ysf - 16
    a = 500 * (xsf - ysf)
    b = 200 * (ysf - zsf)

    return np.array([l, a, b])


def xyz2lab_wo_mon(WP, XYZ):
    xyz_s = XYZ / WP
    xsf = lab_f(xyz_s[0])
    ysf = lab_f(xyz_s[1])
    zsf = lab_f(xyz_s[2])

    l = 116 * ysf - 16
    a = 500 * (xsf - ysf)
    b = 200 * (ysf - zsf)

    return np.array([l, a, b])


def lab_inv_f(t):
    if t > lab_delta:
        return np.power(t, 3)
    else:
        return (3 * (np.power(lab_delta, 2))) * (t - 4/29)


def lab2xyz(mon: Monitor, LAB):
    """
    Convert a color from CIELAB color space to XYZ color space.

    Args:
        mon (Monitor): The monitor object containing the white point values.
        LAB (tuple): A tuple containing the L, a, and b values of the color.

    Returns:
        numpy.ndarray: An array containing the X, Y, and Z values of the color in the XYZ color space.
    """
    l = LAB[0]
    a = LAB[1]
    b = LAB[2]

    x_pre_f = (l+16)/116 + (a/500)
    y_pre_f = (l+16)/116
    z_pre_f = (l+16)/116 - b/200

    X = mon.monWP[0] * lab_inv_f(x_pre_f)
    Y = mon.monWP[1] * lab_inv_f(y_pre_f)
    Z = mon.monWP[2] * lab_inv_f(z_pre_f)

    return np.array([X, Y, Z])


def rgb2lab(mon: Monitor, rgb):
    """
    Convert an RGB color to the CIELAB color space.

    Args:
        mon (Monitor): The monitor object.
        rgb (tuple): A tuple containing the red, green, and blue values of the color.

    Returns:
        tuple: A tuple containing the L*, a*, and b* values of the color in the CIELAB color space.
    """
    xyz = rgb2xyz(mon, rgb)
    return xyz2lab(mon, xyz)


def lab2rgb(mon: Monitor, lab):
    """
    Convert a color from CIELAB color space to RGB color space.

    Args:
        mon (Monitor): The monitor object.
        lab (tuple): The CIELAB color values.

    Returns:
        tuple: The RGB color values.
    """
    xyz = lab2xyz(mon, lab)
    return xyz2rgb(mon, xyz)


def matlab_rgb2lab(mon: Monitor, rgb):
    xyz = matlab_rgb2xyz(mon, rgb)
    return xyz2lab(mon, xyz)


def lstar(mon: Monitor, rgb):
    """
    Convert an RGB color to the L* component of the CIELAB color space.

    Args:
        mon (Monitor): The monitor on which the color is displayed.
        rgb (tuple): A tuple containing the red, green, and blue values of the color.

    Returns:
        float: The L* component of the color in the CIELAB color space.
    """
    lab = rgb2lab(mon, rgb)
    return lab[0]


def chroma(mon: Monitor, rgb):
    """
    Calculates the chroma value of an RGB color using the CIELAB color space.

    Args:
        mon (Monitor): The monitor object used to convert the RGB color to CIELAB.
        rgb (tuple): A tuple containing the red, green, and blue values of the color.

    Returns:
        float: The chroma value of the color.
    """
    lab = rgb2lab(mon, rgb)
    a = lab[1]
    b = lab[2]
    return np.sqrt(np.power(a, 2) + np.power(b, 2))


def hue(lab):
    """
    Calculates the hue value of a color in the CIELAB color space.

    Args:
        lab (tuple): A tuple containing the L*, a*, and b* values of the color.

    Returns:
        float: The hue value of the color in radians.
    """
    a = lab[1]
    b = lab[2]

    h = np.arctan2(b, a)
    if h < 0:
        return h + 2*np.pi

    return h


import numpy as np

from .cie_standard import xyY2XYZ

n = 1

NT_unique_red = 20.14 # degrees
NT_unique_yellow = 90.00 # degrees
NT_unique_green = 164.25 # degrees
NT_unique_blue = 231.00 # degrees

# Hue Quadrature is 400-step hue scale, starting at unique red.
# Interpolate hue based on unique hue angles above to get
# the Hue Quadrature
quad_steps = np.linspace(0, 400)
quad_red = 0
quad_yellow = 100
quad_green = 200
quad_blue = 300

# Hue Composition is what percentage of the arc between two unique
# hues a given color occupies, such as 33Y 67R.
# Interpolate quadrature values to get the percentages for
# Hue Composition.

def eR(R, xi):
    if R >= 20 * xi:
        return 1.758
    else:
        return 1
    

def eG(G, eta):
    if G >= 20 * eta:
        return 1.758
    else:
        return 1


def beta_1(x):
    return (6.496 + 6.362 * np.power(x, 0.4495))/(6.469 + np.power(x, 0.4495))


def beta_2(x):
    return  0.7844 * (8.414 + 8.091 * np.power(x, 0.5128))/(8.414 + np.power(x, 0.5128)) # in Fairchild book


def Es(t):
    return 0.9394 - 0.2478 * np.sin(t) - 0.0743 * np.sin(2*t) + 0.0666 * np.sin(3*t) \
        - 0.0186 * np.sin(4*t) - 0.0055 * np.cos(t) - 0.0521 * np.cos(2*t) \
        - 0.0573 * np.cos(3*t) - 0.0061 * np.cos(4*t)


def robNT(xy_stim, Y_stim, Yo_bg, xy_illum, Eo, Eor):
    """
    Computes the Nayatani et al. (1995) color appearance model.

    Args:
        xy_stim (tuple): Tuple containing the chromaticity coordinates of the stimulus in CIE 1931 xyY color space.
        Y_stim (float): Luminance of the stimulus in cd/m^2.
        Yo_bg (float): Luminance of the background in cd/m^2.
        xy_illum (tuple): Tuple containing the chromaticity coordinates of the illuminant in CIE 1931 xyY color space.
        Eo (float): Illuminance of the stimulus in lux.
        Eor (float): Illuminance of the reference white in lux.

    Returns:
        dict: A dictionary containing the following keys:
            - 'Brightness': Brightness of the stimulus.
            - 'Lightness': Lightness of the stimulus.
            - 'Normalized Lightness': Normalized lightness of the stimulus.
            - 'Hue Angle': Hue angle of the stimulus in degrees.
            - 'Saturation': Saturation of the stimulus.
            - 'Chroma': Chroma of the stimulus.
            - 'Colorfulness': Colorfulness of the stimulus.
            - 'Achromatic Response': Achromatic response of the stimulus.
            - 'RG Response': Red-green response of the stimulus.
            - 'YB Response': Yellow-blue response of the stimulus.
    """
    xo = xy_illum[0]
    yo = xy_illum[1]
    Yo = Yo_bg

    Lo_adapt = (Yo * Eo)/(100 * np.pi)
    Lo_r = (Yo * Eor)/(100 * np.pi)

    xi = (0.48105*xo + 0.78841*yo - 0.08081)/yo
    eta = (-0.272*xo + 1.11962*yo + 0.0457)/yo
    zeta = 0.91822 * (1.0 - xo - yo)/yo

    cone_adapt_resp = Lo_adapt * np.array([xi, eta, zeta])
    R_o = cone_adapt_resp[0]
    G_o = cone_adapt_resp[1]
    B_o = cone_adapt_resp[2]

    xyY_stim = np.array([xy_stim[0], xy_stim[1], Y_stim*Eo])
    XYZ = xyY2XYZ(xyY_stim)
    M = np.array([
        [ 0.40024, 0.7076, -0.08081],
        [-0.2263,  1.16532, 0.0457],
        [0,        0,       0.91822],
    ])
    cone_resp = M @ XYZ
    R = cone_resp[0]
    G = cone_resp[1]
    B = cone_resp[2]

    e_R = eR(R, xi)
    e_G = eG(G, eta)

    b1_Lor = beta_1(Lo_r)
    b1_Ro = beta_1(R_o)
    b1_Go = beta_1(G_o)
    b2_Bo = beta_2(B_o)

    a = 41.69/b1_Lor
    b = (2/3) * b1_Ro * e_R * np.log10((R + n)/(20*xi + n))
    c = (1/3) * b1_Go * e_G * np.log10((G + n)/(20*eta + n))
    Q = a * (b + c)

    d = 50/b1_Lor
    e = (2/3) * b1_Ro
    f = (1/3) * b1_Go
    Br = Q + d*(e + f)

    h = (2/3) * b1_Ro * 1.758 * np.log10((100*xi + n)/(20*xi + n))
    i = (1/3) * b1_Go * 1.758 * np.log10((100*eta + n)/(20*eta + n))
    Brw = a*(h + i) + d*(e + f)

    Lp_star = Q + 50
    Ln_star = 100 * Br/Brw

    ta = b1_Ro * np.log10((R + n)/(20*xi + n))
    tb = (12/11) * b1_Go * np.log10((G + n)/(20*eta + n))
    tc = (1/11) * b2_Bo * np.log10((B + n)/(20*zeta + n))
    t = ta - tb + tc

    pa = (1/9) * b1_Ro * np.log10((R + n)/(20*xi + n))
    pb = (1/9) * b1_Go * np.log10((G + n)/(20*eta + n))
    pc = (2/9) * b2_Bo * np.log10((B + n)/(20*zeta + n))
    p = pa + pb - pc

    theta = np.arctan2(p, t)
    if theta < 0:
        theta += 2*np.pi
    
    thetad = np.rad2deg(theta)

    Es_theta = Es(theta)
    Srg = (488.93 * Es_theta * t)/b1_Lor
    Syb = (488.93 * Es_theta * p)/b1_Lor
    S = np.sqrt(np.power(Srg, 2) + np.power(Syb, 2))

    Lp_scaled = np.power(Lp_star/50, 0.7)
    Crg = Lp_scaled * Srg
    Cyb = Lp_scaled * Syb
    C = Lp_scaled * S

    Mrg = Crg * Brw/100
    Myb = Cyb * Brw/100
    M = C * Brw/100

    NT = {
        'Brightness': Br,
        'Lightness': Lp_star,
        'Normalized Lightness': Ln_star,
        'Hue Angle': thetad,
        'Saturation': S,
        'Chroma': C,
        'Colorfulness': M,
        'Achromatic Response': Q,
        'RG Response': t,
        'YB Response': p,
    }

    return NT


import numpy as np

from .monitor import Monitor

def dkl2rgb(mon: Monitor, ldrgyv):
    """
    Convert a vector of DKL values to RGB values.

    Args:
        mon (Monitor): The monitor object.
        ldrgyv (numpy.ndarray): A vector of DKL values.

    Returns:
        numpy.ndarray: A vector of RGB values.
    """
    return mon.DKL2RGB @ (ldrgyv/2.0 + 0.5)


def rgb2dkl(mon: Monitor, rgb):
    """
    Convert an RGB color to DKL color space.

    Args:
        mon (Monitor): The monitor object containing the DKL2RGB matrix.
        rgb (numpy.ndarray): The RGB color to convert.

    Returns:
        numpy.ndarray: The DKL color space representation of the input RGB color.
    """
    rgbScaled = 2.0 * (rgb - 0.5)
    return np.linalg.inv(mon.DKL2RGB) @ rgbScaled


import matplotlib.pyplot as plt
from matplotlib import ticker
import numpy as np

from .monitor import Monitor
from .cie_standard import plank_lasso
from .uv_space import plank_lasso_uvY
from .infamous_lab import lab2rgb

def xyY_fig(mon: Monitor):
    fig, ax = plt.subplots(1, 1, figsize = (7, 7))
    plt.grid()

    ax.plot(plank_lasso[:, 0], plank_lasso[:, 1], color="black")

    ax.plot([mon.monxyY[0, 0], mon.monxyY[1, 0]], [mon.monxyY[0, 1], mon.monxyY[1, 1]], color="black")
    ax.plot([mon.monxyY[1, 0], mon.monxyY[2, 0]], [mon.monxyY[1, 1], mon.monxyY[2, 1]], color="black")
    ax.plot([mon.monxyY[0, 0], mon.monxyY[2, 0]], [mon.monxyY[0, 1], mon.monxyY[2, 1]], color="black")
    ax.scatter(mon.monxyY[0, 0], mon.monxyY[0, 1], color="red")
    ax.scatter(mon.monxyY[1, 0], mon.monxyY[1, 1], color="green")
    ax.scatter(mon.monxyY[2, 0], mon.monxyY[2, 1], color="blue")

    ax.set_xlim((0, 0.8))
    ax.set_ylim((0, 0.9))

    ax.set_xlabel('x')
    ax.set_ylabel('y')

    return fig, ax


def AB_fig(mon: Monitor, xlims, ylims):
    fig, ax = plt.subplots(1, 1, figsize = (7, 7))
    plt.grid()

    plt.plot(np.array([0, 0]), np.array([ylims[0], ylims[1]]), color="black")
    plt.plot(np.array([xlims[0], xlims[1]]), np.array([0, 0]), color="black")

    ax.set_xlim(xlims)
    ax.set_ylim(ylims)

    ax.set_xlabel('a*')
    ax.set_ylabel('b*')

    return fig, ax


def LC_fig(mon: Monitor, xlims, ylims):
    fig, ax = plt.subplots(1, 1, figsize = (7, 7))
    plt.grid()

    ax.set_xlim(xlims)
    ax.set_ylim(ylims)

    ax.set_xlabel('Chroma')
    ax.set_ylabel('L*')

    return fig, ax


def uvY_fig(mon: Monitor):
    fig, ax = plt.subplots(1, 1, figsize = (7, 7))
    plt.grid()

    ax.plot(plank_lasso_uvY[:, 0], plank_lasso_uvY[:, 1], color="black")

    # ax.plot([mon.monuvY[0, 0], mon.monuvY[1, 0]], [mon.monuvY[0, 1], mon.monuvY[1, 1]], color="black", linestyle="dashed")
    # ax.plot([mon.monuvY[1, 0], mon.monuvY[2, 0]], [mon.monuvY[1, 1], mon.monuvY[2, 1]], color="black", linestyle="dashed")
    # ax.plot([mon.monuvY[0, 0], mon.monuvY[2, 0]], [mon.monuvY[0, 1], mon.monuvY[2, 1]], color="black", linestyle="dashed")
    # ax.scatter(mon.monuvY[0, 0], mon.monuvY[0, 1], color="red")
    # ax.scatter(mon.monuvY[1, 0], mon.monuvY[1, 1], color="green")
    # ax.scatter(mon.monuvY[2, 0], mon.monuvY[2, 1], color="blue")

    ax.set_xlim((0, 0.6))
    ax.set_ylim((0, 0.7))

    ax.set_xlabel('u\'')
    ax.set_ylabel('v\'')

    return fig, ax


def add_hue_plot(mon: Monitor, fig, subspec):
    position_major = [0, np.pi/6, np.pi/3, np.pi/2, 3*np.pi/4, np.pi, 5*np.pi/4, 3*np.pi/2, 7*np.pi/4]
    labels = ["Red", "Pink", "Orange", "Yellow", "Chartreuse", "Green", "Cyan", "Blue", "Purple"]

    # rgbs = np.zeros((len(labels), 3))
    # for c in range(len(labels)):
        # l = 50
        # a = 80*np.cos(position_major[c])
        # b = 80*np.sin(position_major[c])
        # rgb = lab2rgb(mon, np.array([l, a, b]))
        # rgbs[c, :] = np.sqrt(rgb)

    ax = fig.add_subplot(subspec, projection='polar')
    ax.grid(True)
    ax.axis(True)
    ax.spines["polar"].set_visible(False)
    ax.xaxis.set_major_locator(ticker.FixedLocator(position_major))
    ax.xaxis.set_major_formatter(ticker.FixedFormatter(labels))
    ax.set_rgrids([0, 1], ["" for _ in range(len([0, 1]))], fontsize=16)
    ax.set_ylim(0, 1.12)
    ax.set_yticklabels([])

    # ax.scatter(position_major, np.ones((len(position_major))), c=rgbs, s=80)

    return ax


import numpy as np

# what happens when using L+M light to stimulate
# a 2L+M cell.

from .monitor import Monitor

def rgb2lms(mon: Monitor, rgb):
    """
    Convert RGB values to LMS values using the given monitor calibration.

    Args:
        mon (Monitor): The monitor calibration to use for the conversion.
        rgb (array-like): The RGB values to convert.

    Returns:
        array-like: The LMS values corresponding to the input RGB values.
    """
    return mon.RGB2LMS @ rgb


# sources:
# http://cvrl.ucl.ac.uk/database/text/cienewxyz/cie2012xyz2.htm
# https://en.wikipedia.org/wiki/LMS_color_space#cite_note-13
def xyz2lms(xyz):
    """
    Convert from CIE XYZ color space to LMS color space using the CIE 2006 LMS to XYZ matrix.

    Parameters:
    -----------
    xyz : numpy.ndarray
        A 3-element array representing the XYZ color values.

    Returns:
    --------
    numpy.ndarray
        A 3-element array representing the LMS color values.
    """
    cie2006_lms2xyz = np.array([[1.94735469, -1.41445123, 0.36476327],
                                [0.68990272,  0.34832189, 0],
                                [0,           0,          1.93485343]])

    xyz2lms_mat = np.linalg.inv(cie2006_lms2xyz)
    return xyz2lms_mat @ xyz


def lms2dkl(lms):
    """
    Convert LMS color space to DKL color space.

    Parameters:
    -----------
    lms : numpy.ndarray
        Array of LMS color values.

    Returns:
    --------
    numpy.ndarray
        Array of DKL color values.
    """
    l = lms[:, 0].squeeze()
    m = lms[:, 1].squeeze()
    s = lms[:, 2].squeeze()

    ld = l + m
    rg = l - m
    yv = 2.0*s - ld

    dkl = np.zeros(lms.shape)
    dkl[:, 0] = ld
    dkl[:, 1] = rg
    dkl[:, 2] = yv
    
    return dkl


def maxLMS(lms):
    return np.max(lms)


def maxWeightLMS(lms_s, wL, wM, wS):
    w_lms_s = np.array([[wL], [wM], [wS]]) * lms_s
    return np.array(np.max(w_lms_s, axis=0)).flatten()


# TODO: get weights from Shuchen
def maxLMS_Shuchen(lms):
    ws = np.array([0.40, 0.45, 0.15])
    return np.max(ws * lms)


rand_weights = np.random.rand(1, 3)
def maxLMS_Random(lms):
    return np.max(rand_weights * lms)



import numpy as np

from .monitor import Monitor

def rank_lum(mon: Monitor, rgbs):
    """
    Ranks the brightness of RGB colors based on the luminance values of the monitor.

    Args:
        mon (Monitor): The monitor object containing the luminance values.
        rgbs (numpy.ndarray): An array of RGB colors to be ranked.

    Returns:
        numpy.ndarray: An array of indices that sorts the RGB colors from highest to lowest brightness.
    """
    rs = rgbs[:, 0].squeeze()
    gs = rgbs[:, 1].squeeze()
    bs = rgbs[:, 2].squeeze()

    rY = mon.monxyY[0, 2]
    gY = mon.monxyY[1, 2]
    bY = mon.monxyY[2, 2]

    lums = rs*rY + gs*gY + bs*bY
    rnks = np.argsort(lums) # sorts from lowest to highest
    return np.flip(rnks)


def rank_rad(mon: Monitor, rgbs):
    """
    Ranks the brightness of RGB values based on the maximum spectrum of the given monitor.

    Args:
        mon (Monitor): The monitor object containing the maximum spectrum values.
        rgbs (numpy.ndarray): An array of RGB values.

    Returns:
        numpy.ndarray: An array of indices that correspond to the brightness ranking of the RGB values.
    """
    rs = rgbs[:, 0].squeeze()
    gs = rgbs[:, 1].squeeze()
    bs = rgbs[:, 2].squeeze()

    rspec = mon.R_max_spectrum
    gspec = mon.G_max_spectrum
    bspec = mon.B_max_spectrum

    rads = np.zeros((len(rs), ))
    for c in range(len(rs)):
        r_rad = (rs[c]*rspec).sum()
        g_rad = (gs[c]*gspec).sum()
        b_rad = (bs[c]*bspec).sum()

        rads[c] = r_rad + g_rad + b_rad

    rnks = np.argsort(rads) # sorts from lowest to highest
    return np.flip(rnks)


def rank_sumrgb(rgbs):
    """
    Ranks the brightness of RGB values based on the sum of their components.

    Args:
        rgbs (numpy.ndarray): An array of RGB values, where each row represents a single RGB value.

    Returns:
        numpy.ndarray: An array of indices that represent the ranking of the RGB values based on their brightness.
    """
    rs = rgbs[:, 0].squeeze()
    gs = rgbs[:, 1].squeeze()
    bs = rgbs[:, 2].squeeze()

    ms = np.zeros((len(rs), ))
    for c in range(len(rs)):
        ms[c] = rs[c] + gs[c] + bs[c]

    rnks = np.argsort(ms) # sorts from lowest to highest
    return np.flip(rnks)


def rank_maxrgb(rgbs):
    """
    Ranks the RGB values based on their maximum value.

    Args:
        rgbs (numpy.ndarray): An array of RGB values.

    Returns:
        numpy.ndarray: An array of indices that correspond to the sorted RGB values in descending order.
    """
    rs = rgbs[:, 0].squeeze()
    gs = rgbs[:, 1].squeeze()
    bs = rgbs[:, 2].squeeze()

    ms = np.zeros((len(rs), ))
    for c in range(len(rs)):
        ms[c] = np.max(np.array([rs[c], gs[c], bs[c]]))

    rnks = np.argsort(ms) # sorts from lowest to highest
    return np.flip(rnks)


def normalised_kendall_tau_distance(values1, values2):
    """
    Compute the normalised Kendall tau distance between two lists of values.

    Args:
        values1 (list): First list of values.
        values2 (list): Second list of values.

    Returns:
        float: Normalised Kendall tau distance between the two lists of values.
    """
    n = len(values1)
    assert len(values2) == n, "Both lists have to be of equal length"
    i, j = np.meshgrid(np.arange(n), np.arange(n))
    a = np.argsort(values1)
    b = np.argsort(values2)
    ndisordered = np.logical_or(np.logical_and(a[i] < a[j], b[i] > b[j]), np.logical_and(a[i] > a[j], b[i] < b[j])).sum()
    return ndisordered / (n * (n - 1) / 2.0) # corrected based on wikipedia article and c code


def accuracyShuchen(ranks1, ranks2):
    """
    Computes the accuracy of the ranking given two sets of ranks.

    Parameters:
    ranks1 (list): A list of ranks.
    ranks2 (list): A list of ranks.

    Returns:
    float: The accuracy of the ranking.
    """
    nr1 = len(ranks1)
    pairRanks1 = np.zeros((nr1, nr1))
    for c1 in range(nr1):
        for cc1 in range(nr1):
            rl = ranks1[c1]
            rr = ranks1[cc1]
            if rl > rr:
                pairRanks1[c1, cc1] = 1
            else:
                pairRanks1[c1, cc1] = 0
    
    nr2 = len(ranks2)
    pairRanks2 = np.zeros((nr2, nr2))
    for c1 in range(nr2):
        for cc1 in range(nr2):
            rl = ranks2[c1]
            rr = ranks2[cc1]
            if rl > rr:
                pairRanks2[c1, cc1] = 1
            else:
                pairRanks2[c1, cc1] = 0

    acc = (pairRanks1 == pairRanks2).flatten().sum()

    return acc/(nr1*nr1)

from .maxRGB_revolution import maxRGB, maxRGB_Shuchen, maxRGB_G0
from .sumrgb import sumRGB_Shuchen, sumRGB
from .ware_cowan import ware_cowan
from .fairchildL import rgb2FairchildLstar
from .funny_dude_discovers_maxRGB_in_adobe import RGBtoHSP
from .G0_brilliance import brilliance, brilliance_macadam
from .radiance import radiance, visibleRadiance
from .cie_monitor_helpers import rgb2xyz, luminance, xyY2rgb
from .cie_standard import XYZ2xyY, XYZ2xy
from .lms import rgb2lms, maxLMS
from .infamous_lab import rgb2lab, chroma
from .guth_vector import guth_lum_rgb
from .jazzy import jaz_lightness, jazzy
from .nayatani_insanity import robNT
from .hunt import hunt_brightness
from .svf import svf
from .jingHSP import jingHSP

import colour

def computeAllBrightnessModels(mon: Monitor, rgbs):
    """
    Computes various brightness models for a given set of RGB values.

    Args:
    - mon: Monitor object containing information about the monitor used to display the RGB values.
    - rgbs: numpy array of shape (n, 3) containing n RGB values.

    Returns:
    - Dictionary containing the computed brightness values for each model.
    """
    d65_xy = XYZ2xy(mon.monWP)

    ncs = len(rgbs)

    brightnesses = {
        'maxRGB': np.zeros((ncs)),
        'maxRGB_Shuchen': np.zeros((ncs)),
        'maxRGB_G0': np.zeros((ncs)),
        'sumRGB': np.zeros((ncs)),
        'sumRGB_Shuchen': np.zeros((ncs)),
        'ware_cowan': np.zeros((ncs)),
        'Lstar': np.zeros((ncs)),
        'fairchildL': np.zeros((ncs)),
        'HSP': np.zeros((ncs)),
        'HSP_JingWeights': np.zeros((ncs)),
        'maxLMS': np.zeros((ncs)),
        'brilliance': np.zeros((ncs)),
        'brilliance_desat': np.zeros((ncs)),
        'brill_macadam': np.zeros((ncs)),
        'luminance': np.zeros((ncs)),
        'radiance': np.zeros((ncs)),
        'chroma': np.zeros((ncs)),
        'visibleRadiance': np.zeros((ncs)),
        'Nayatani Br': np.zeros((ncs)),
        'Nayatani Lp': np.zeros((ncs)),
        'Nayatani Ln': np.zeros((ncs)),
        'CAM16 Avg Q': np.zeros((ncs)),
        'CAM16 Avg J': np.zeros((ncs)),
        'CAM16 Dim Q': np.zeros((ncs)),
        'CAM16 Dim J': np.zeros((ncs)),
        'CAM16 Dark Q': np.zeros((ncs)),
        'CAM16 Dark J': np.zeros((ncs)),
        'Guth': np.zeros((ncs)),
        'JzAzBz': np.zeros((ncs)),
        'Hunt_Q Tele': np.zeros((ncs)),
        'Hunt_WB Tele': np.zeros((ncs)),
        'Hunt_Q Norm': np.zeros((ncs)),
        'Hunt_WB Norm': np.zeros((ncs)),
        'Hunt_Q Small': np.zeros((ncs)),
        'Hunt_WB Small': np.zeros((ncs)),
        'JingB': np.zeros((ncs)),
        'ATD': np.zeros((ncs)),
        'Hellwig Avg Q': np.zeros((ncs)),
        'Hellwig Avg J': np.zeros((ncs)),
        'Hellwig Dim Q': np.zeros((ncs)),
        'Hellwig Dim J': np.zeros((ncs)),
        'Hellwig Dark Q': np.zeros((ncs)),
        'Hellwig Dark J': np.zeros((ncs)),
        'SVF': np.zeros((ncs))
    }

    for c in range(ncs):
        rgb = rgbs[c, :].squeeze()

        r = rgb[0]
        g = rgb[1]
        b = rgb[2]

        xyz = rgb2xyz(mon, rgb)
        xyY = XYZ2xyY(xyz)

        xy_desat = 0.4*xyY[0:2] + 0.6*d65_xy

        lms = rgb2lms(mon, rgb)

        lab = rgb2lab(mon, rgb)
        brightnesses['Lstar'][c] = lab[0]

        brightnesses['maxRGB'][c] = maxRGB(rgb)
        brightnesses['maxRGB_Shuchen'][c] = maxRGB_Shuchen(rgb)
        brightnesses['maxRGB_G0'][c] = maxRGB_G0(rgb)

        brightnesses['sumRGB'][c] = sumRGB(rgb)
        brightnesses['sumRGB_Shuchen'][c] = sumRGB_Shuchen(rgb)

        brightnesses['ware_cowan'][c] = ware_cowan(xyY)
        brightnesses['fairchildL'][c] = rgb2FairchildLstar(mon, rgb)

        hsp = RGBtoHSP(r, g, b, "photoshop")
        brightnesses['HSP'][c] = hsp[2]

        hsp = RGBtoHSP(r, g, b, "jing")
        brightnesses['HSP_JingWeights'][c] = hsp[2]

        brightnesses['JingB'][c] = jingHSP(rgb)

        brightnesses['maxLMS'][c] = maxLMS(lms)

        brightnesses['brilliance'][c] = brilliance(mon, rgb)

        xyY_desat = np.array([xy_desat[0], xy_desat[1], xyY[2]])
        rgb_desat = xyY2rgb(mon, xyY_desat)
        brightnesses['brilliance_desat'][c] = brilliance(mon, rgb_desat)

        brightnesses['brill_macadam'][c] = brilliance_macadam(mon, rgb)

        brightnesses['luminance'][c] = luminance(mon, rgb)
        brightnesses['radiance'][c] = radiance(mon, rgb)
        brightnesses['chroma'][c] = chroma(mon, rgb)

        brightnesses['visibleRadiance'][c] = visibleRadiance(mon, rgb)

        brightnesses['Guth'][c] = guth_lum_rgb(mon, rgb)

        jaz = jazzy(mon, rgb)
        brightnesses['JzAzBz'][c] = jaz_lightness(jaz)

        Eo = mon.monWP[1]
        Eor = 1000
        Y_stim = xyY[2]/Eo
        Yo_bg = 0 # we always had a black background in all experiments
        xy_illum = XYZ2xy(mon.monWP)
        xy_stim = xyY[0:2]
        nt_res = robNT(xy_stim, Y_stim, Yo_bg, xy_illum, Eo, Eor)
        brightnesses['Nayatani Br'][c] = nt_res['Brightness']
        brightnesses['Nayatani Lp'][c] = nt_res['Lightness']
        brightnesses['Nayatani Ln'][c] = nt_res['Normalized Lightness']

        XYZ_w = mon.monWP
        L_A = 0.2*mon.monWP[1]
        Y_b = 0.001
        surround = colour.appearance.VIEWING_CONDITIONS_CAM16["Average"]
        cam16 = colour.appearance.XYZ_to_CAM16(xyz, XYZ_w, L_A, Y_b, surround) 
        brightnesses['CAM16 Avg Q'][c] = cam16.Q
        brightnesses['CAM16 Avg J'][c] = cam16.J

        surround = colour.appearance.VIEWING_CONDITIONS_CAM16["Dim"]
        cam16 = colour.appearance.XYZ_to_CAM16(xyz, XYZ_w, L_A, Y_b, surround) 
        brightnesses['CAM16 Dim Q'][c] = cam16.Q
        brightnesses['CAM16 Dim J'][c] = cam16.J

        surround = colour.appearance.VIEWING_CONDITIONS_CAM16["Dark"]
        cam16 = colour.appearance.XYZ_to_CAM16(xyz, XYZ_w, L_A, Y_b, surround) 
        brightnesses['CAM16 Dark Q'][c] = cam16.Q
        brightnesses['CAM16 Dark J'][c] = cam16.J

        # values for Nc, Nb
        # small area, uniform surround = 1.0, 300
        # normal scenes = 1.0, 75
        # television, dim surround = 1.0, 25
        # transparencies, light box = 0.7, 25
        # transparencies, dark surround = 0.7, 10
        xy_adapt = np.array([0.33, 0.33])
        Yref_adapt = 0.001
        xy_bkgd = np.array([0.33, 0.33])
        Yref_bkgd = 0.001
        Nc = 1.0
        Nb = 25
        hunt_res = hunt_brightness(xy_stim, Y_stim, xy_illum, Eo, xy_adapt, Yref_adapt, xy_bkgd, Yref_bkgd, Nc, Nb)
        brightnesses['Hunt_Q Tele'] = hunt_res['Q']
        brightnesses['Hunt_WB Tele'] = hunt_res['WB']

        Nc = 1.0
        Nb = 75
        hunt_res = hunt_brightness(xy_stim, Y_stim, xy_illum, Eo, xy_adapt, Yref_adapt, xy_bkgd, Yref_bkgd, Nc, Nb)
        brightnesses['Hunt_Q Norm'] = hunt_res['Q']
        brightnesses['Hunt_WB Norm'] = hunt_res['WB']

        Nc = 1.0
        Nb = 300
        hunt_res = hunt_brightness(xy_stim, Y_stim, xy_illum, Eo, xy_adapt, Yref_adapt, xy_bkgd, Yref_bkgd, Nc, Nb)
        brightnesses['Hunt_Q Small'] = hunt_res['Q']
        brightnesses['Hunt_WB Small'] = hunt_res['WB']

        # we were viewing in related colors mode
        atd_k_1 = 1.0
        atd_k_2 = 5.0
        atd_res = colour.appearance.atd95.XYZ_to_ATD95(xyz, mon.monWP, Yref_adapt, atd_k_1, atd_k_2)
        brightnesses['ATD'][c] = atd_res.Q

        surr = colour.appearance.hellwig2022.VIEWING_CONDITIONS_HELLWIG2022["Dim"]
        hf_res = colour.appearance.XYZ_to_Hellwig2022(xyz, mon.monWP, Yref_adapt, Yref_bkgd, surr)
        brightnesses['Hellwig Dim Q'][c] = hf_res.Q
        brightnesses['Hellwig Dim J'][c] = hf_res.J

        surr = colour.appearance.hellwig2022.VIEWING_CONDITIONS_HELLWIG2022["Average"]
        hf_res = colour.appearance.XYZ_to_Hellwig2022(xyz, mon.monWP, Yref_adapt, Yref_bkgd, surr)
        brightnesses['Hellwig Avg Q'][c] = hf_res.Q
        brightnesses['Hellwig Avg J'][c] = hf_res.J

        surr = colour.appearance.hellwig2022.VIEWING_CONDITIONS_HELLWIG2022["Dark"]
        hf_res = colour.appearance.XYZ_to_Hellwig2022(xyz, mon.monWP, Yref_adapt, Yref_bkgd, surr)
        brightnesses['Hellwig Dark Q'][c] = hf_res.Q
        brightnesses['Hellwig Dark J'][c] = hf_res.J

        svf_res = svf(xyz, mon.monWP)
        brightnesses['SVF'][c] = svf_res[2]

    return brightnesses


import numpy as np

# Converts xyY to the Kaiser L** measure for perceived brightness
def ware_cowan(xyY):
    x = xyY[0]
    y = xyY[1]
    Y = xyY[2]

    a = 0.184 * y
    b = 2.527 * x * y
    c = 4.65 * np.power(x, 3) * y
    d = 4.657 * x * np.power(y, 4)

    F = 0.256 - a - b + c + d
    
    return np.log10(Y) + F


import numpy as np

def atan2pi_to_normal_pi(x):
    """
    Converts an angle in the range [-pi, pi] to the range [0, 2*pi].

    Args:
        x (float): Angle in radians.

    Returns:
        float: Angle in radians in the range [0, 2*pi].
    """
    if x < 0:
        return np.mod(x, 2*np.pi)
    else:
        return x



import numpy as np

from .monitor import Monitor

def radiance(mon: Monitor, rgb):
    """
    Calculates the radiance of a given RGB value using the provided monitor object.

    Args:
        mon (Monitor): The monitor object containing the R, G, and B max spectra and max radiance values.
        rgb (tuple): A tuple containing the RGB values to calculate radiance for.

    Returns:
        float: The calculated radiance value.
    """
    r = rgb[0]
    g = rgb[1]
    b = rgb[2]

    return (mon.R_max_spectrum*r + mon.G_max_spectrum*g + mon.B_max_spectrum*b).sum()/mon.maxRadiance


def visibleRadiance(mon: Monitor, rgb):
    """
    Calculates the visible radiance of a given RGB value on a given monitor.

    Args:
        mon (Monitor): The monitor object.
        rgb (tuple): The RGB value to calculate the visible radiance for.

    Returns:
        float: The visible radiance of the given RGB value on the given monitor.
    """
    r = rgb[0]
    g = rgb[1]
    b = rgb[2]

    return (mon.visible_R_max_spectrum*r + mon.visible_G_max_spectrum*g + mon.visible_B_max_spectrum*b).sum()/mon.maxVisibleRadiance


import numpy as np
import colour

from .hunt import hunt_brightness
from .cie_standard import XYZ2xy

# Table 12.3 Example Hunt color appearance model calculations
# Quantity Case 1 Case 2 Case 3 Case 4
Xs = [19.01, 57.06, 3.53, 19.01]
Ys = [20.00, 43.06, 6.56, 20.00]
Zs = [21.78, 31.96, 2.14, 21.78]
XWs = [95.05, 95.05, 109.85, 109.85]
YWs = [100.00, 100.00, 100.00, 100.00]
ZWs = [108.88, 108.88, 35.58, 35.58]
LAs = [318.31, 31.83, 318.31, 31.83]
Ncs = [1.0, 1.0, 1.0, 1.0]
Nbs = [75, 75, 75, 75]
Discounting = [True, True, True, True]
hSs = [269.3, 18.6, 178.3, 262.8]
Hs = [317.2, 398.8, 222.2, 313.4]
HCs = ["83B 17R", "99R 1B", "78G 22B", "87B 13R"]
ss = [0.03, 153.36, 245.40, 209.29]
Qs = [31.92, 31.22, 18.90, 22.15]
Js = [42.12, 66.76, 19.56, 40.27]
C94s = [0.16, 63.89, 74.58, 73.84]
M94s = [0.16, 58.28, 76.33, 67.35]

for tc in range(4):
    X = Xs[tc]
    Y = Ys[tc]
    Z = Zs[tc]

    XW = XWs[tc]
    YW = YWs[tc]
    ZW = ZWs[tc]

    Nc = Ncs[tc]
    Nb = Nbs[tc]

    xy_illum = XYZ2xy(np.array([XW, YW, ZW]))
    Y_illum = LAs[tc]
    
    xy_stim = XYZ2xy(np.array([X, Y, Z]))
    Yref_stim = Y/LAs[tc]
    
    # assumptions about background and adapt
    # from Fairchild 
    xy_adapt = XYZ2xy(np.array([XW, YW, ZW]))
    Yref_adapt = 0.20
    
    xy_bkgd = XYZ2xy(np.array([XW, YW, ZW]))
    Yref_bkgd = 0.20

    discount = Discounting[tc]

    XYZ = np.array([X, Y, Z]) 
    XYZ_w = np.array([XW, YW, ZW])
    XYZ_b = np.array([XW, 0.20*YW, ZW])
    L_A = Y_illum
    T = colour.temperature.xy_to_CCT(xy_illum, "Hernandez 1999")

    # res = hunt_brightness(xy_stim, Yref_stim, xy_illum, Y_illum, xy_adapt, Yref_adapt, xy_bkgd, Yref_bkgd, Nc, Nb, discount, False)
    res = hunt_brightness(XYZ, XYZ_w, L_A, np.array([0, 0, 0]), XYZ_b, Nc, Nb, discount, False)
    res_Q = res['Q']
    # res_WB = res['WB']

    res_col = colour.appearance.hunt.XYZ_to_Hunt(XYZ, XYZ_w, XYZ_b, L_A, CCT_w=T, discount_illuminant=discount)
    res_col_Q = res_col.Q

    if Qs[tc] != res_Q:
        print(tc, "Qs: ", Qs[tc], ", Rob: ", res_Q, "Colour: ", res_col_Q)



import numpy as np
from scipy.interpolate import LinearNDInterpolator

from .monitor import Monitor
from .uv_space import uvY2xyY
from .cie_standard import XYZ2xyY
from .macadam import macadam_interp

G0_stims_uvY = np.array([
    [0.448, 0.5024, 72.94], # 1
    [0.3453, 0.4970, 179.31], # 2
    [0.2548, 0.4829, 338.67], # 3
    [0.2333, 0.5556, 361.72], # 4
    [0.2242, 0.5385, 435.82], # 5
    [0.2117, 0.5104, 431.28], # 6
    [0.1201, 0.5092, 334.39], # 7
    [0.1400, 0.5006, 387.97], # 8
    [0.1667, 0.4869, 466.44], # 9
    [0.1415, 0.4180, 310.11], # 10
    [0.1577, 0.4361, 379.57], # 11
    [0.1769, 0.4537, 448.53], # 12
    [0.2537, 0.3307, 181.88], # 13
    [0.2331, 0.3794, 275.61], # 14
    [0.2132, 0.4297, 411.25], # 15
    [0.1978, 0.4683, 530.00]  # 16
])

G0_stims_xyY = np.zeros(G0_stims_uvY.shape)
for c in range(len(G0_stims_xyY)):
    G0_stims_xyY[c, :] = uvY2xyY(G0_stims_uvY[c, :])


xs = G0_stims_xyY[:, 0]
ys = G0_stims_xyY[:, 1]
zs = G0_stims_xyY[:, 2]
brill_interp = LinearNDInterpolator(list(zip(xs, ys)), zs)

def local_rgb2xyz(mon: Monitor, rgb):
    return mon.monXYZ @ rgb

# brilliance according to the function in fig. 4.5 of Xie's thesis:
# https://scholarworks.rit.edu/cgi/viewcontent.cgi?article=12595&context=theses
# page 66 of the PDF file
def brilliance(mon: Monitor, rgb):
    """
    Calculates the brilliance of a given RGB color on a given monitor.

    Args:
        mon (Monitor): The monitor on which the color is displayed.
        rgb (tuple): The RGB color to calculate the brilliance for.

    Returns:
        float: The calculated brilliance value.
    """
    xyz = local_rgb2xyz(mon, rgb)
    xyY = XYZ2xyY(xyz)
    g0_boundary_lum = brill_interp(xyY[0], xyY[1])
    g0_normed_lum = xyY[2]/g0_boundary_lum
    return np.power(g0_normed_lum, 1/2.05)


def brilliance_macadam(mon: Monitor, rgb):
    """
    Calculates the brilliance of a color using the MacAdam method.

    Args:
        mon (Monitor): The monitor object used for calibration.
        rgb (tuple): The RGB values of the color.

    Returns:
        float: The brilliance of the color.
    """
    xyz = local_rgb2xyz(mon, rgb)
    xyY = XYZ2xyY(xyz)

    # in Xie's thesis, they related macadam optimal colors at 200 cd/m2 to G0
    macadam_boundary_lum = macadam_interp(xyY[0], xyY[1])
    g0_normed_lum = xyY[2]/macadam_boundary_lum
    return np.power(g0_normed_lum, 1/2.05)